# coding: utf-8
"""Tools for writing Xcrysden files."""
from __future__ import annotations

import numpy as np

from pymatgen.core.units import Energy, EnergyArray
from abipy.tools.numtools import transpose_last3dims, add_periodic_replicas


__all__ = [
    "xsf_write_structure_and_data_to_path",
    "xsf_write_structure",
    "xsf_write_data",
    "bxsf_write",
]


def xsf_write_structure(file, structures: list) -> None:
    """
    Write the crystalline structure in the Xcrysden format (XSF)

    Args:
        file: file-like object.
        structures: :class:`Structure` or list of :class:`Structure` objects.
    """
    animation = True
    if not isinstance(structures, (list, tuple)):
        structures = [structures]
        animation = False

    fwrite = file.write

    if animation:
        # axsf file.
        fwrite('ANIMSTEPS %s\n' % len(structures))

    fwrite('CRYSTAL\n')

    for n, struct in enumerate(structures):
        cell = struct.lattice.matrix

        fwrite('# Primitive lattice vectors in Angstrom\n')
        fwrite('PRIMVEC %d\n' % (n + 1))
        for i in range(3):
            fwrite(' %.14f %.14f %.14f\n' % tuple(cell[i]))

        cart_coords = struct.cart_coords
        atomic_numbers = struct.atomic_numbers

        # TODO
        cart_forces = None
        #if "cartesian_forces" in structure.site_properties:
        #cart_forces = ArrayWithUnit().to("Ha ang^-1")

        fwrite("# Cartesian coordinates in Angstrom.\n")
        fwrite('PRIMCOORD %d\n' % (n + 1))
        fwrite(' %d 1\n' % len(cart_coords))

        for a in range(len(cart_coords)):
            fwrite(' %2d' % atomic_numbers[a])
            fwrite(' %20.14f %20.14f %20.14f' % tuple(cart_coords[a]))
            if cart_forces is None:
                fwrite('\n')
            else:
                fwrite(' %20.14f %20.14f %20.14f\n' % tuple(cart_forces[a]))


def xsf_write_structure_and_data_to_path(filepath, structure, datar, **kwargs) -> None:
    """Simplified interface to write structure and data to filepath in XSF format."""
    with open(filepath, mode="wt") as fh:
        xsf_write_structure(fh, structure)
        xsf_write_data(fh, structure, datar, **kwargs)


#def xsf_write_structure_and_multidata(filepath, structure, multi_datar, tags=None, **kwargs) -> None:
#    """
#    """
#    multi_datar = np.array(multi_datar, ndmin=4)
#    if tags is not None:
#        if len(tags) != len(multi_datar):
#            raise ValueError(f"tags and multi_datar should have same length but {len(tags)} != {len(multi_datar)}")
#
#    with open(filepath, mode="wt") as fh:
#        xsf_write_structure(fh, structure)
#        for i, datar in enumerate(multi_datar):
#           tag = tags[i] if tags is not None else f"data_{i}"
#           xsf_write_data(fh, structure, multi_datar[i], tag=tag, **kwargs)



def xsf_write_data(file, structure, data, add_replicas=True, cplx_mode=None,
                   idname="data", tag="_UNKNOWN") -> None:
                   #idname="data", tag="_grid") -> None:
    """
    Write data in the Xcrysden format (XSF)

    Args:
        file: file-like object.
        structure: :class:`Structure` object.
        data: array-like object in C-order, i.e data[nx, ny, nz] or data[ngrids, nx, ny, nz]
        add_replicas: If True, data is padded with redundant data points.
            in order to have a periodic 3D array of shape: (nx+1, ny+1, nz+1).
        cplx_mode: string defining the data to print when data is a complex array.
            Possible choices are (case-insensitive):

                - "re"  for real part.
                - "im" for imaginary part.
                - "abs" for the absolute value
    """
    fwrite = file.write

    # Check this one
    if add_replicas:
        data = add_periodic_replicas(data)

    if np.iscomplexobj(data):
        if cplx_mode is None:
            raise TypeError("cplx_mode must be specified when data is a complex array.")
        cplx_mode = cplx_mode.lower()
        if cplx_mode == "re":
            data = data.real
        elif cplx_mode == "im":
            data = data.imag
        elif cplx_mode == "abs":
            data = np.abs(data)
        else:
            raise ValueError("Wrong value for cplx_mode: %s" % cplx_mode)

    shape, ndim = data.shape, data.ndim

    if ndim == 3:
        ngrids = 1
        data = np.asarray([data])
    elif ndim == 4:
        ngrids = shape[0]
    else:
        raise ValueError("ndim %d is not supported" % ndim)

    # Xcrysden uses Fortran-order.
    # Transpose (...,x,y,z) --> (...,z,y,x) to speed up the write below.
    fdata = transpose_last3dims(data)
    fgrid = fdata.shape[-3:]

    cell = structure.lattice.matrix
    origin = np.zeros(3)

    fwrite('BEGIN_BLOCK_DATAGRID_3D\n')
    fwrite(f' {idname}\n')

    for dg in range(ngrids):
        if ngrids != 1:
            fwrite(f" BEGIN_DATAGRID_3D{tag}#{dg+1}" + "\n")
        else:
            fwrite(f" BEGIN_DATAGRID_3D{tag}" + "\n")
        fwrite('%d %d %d\n' % shape[-3:])

        fwrite('%f %f %f\n' % tuple(origin))
        for i in range(3):
            fwrite('%f %f %f\n' % tuple(cell[i]))

        for z in range(fgrid[0]):
            for y in range(fgrid[1]):
                slice_x = fdata[dg,z,y]
                fwrite(' '.join(['%f' % d for d in slice_x]))
                fwrite('\n')
            fwrite('\n')

        fwrite(' END_DATAGRID_3D\n')
    fwrite('END_BLOCK_DATAGRID_3D\n')


def bxsf_write(file, structure, nsppol, nband, ngkpt, ucdata_sbk, fermie, unit="eV") -> None:
    """
    Write band structure data in the Xcrysden format (XSF)

    Args:
        file: file-like object.
        structure: :class:`Structure` object.
        nsppol: Number of spins.
        nband: Number of bands.
        ngkpt: Number of divisions of the full k-mesh.
        ucdata_sbk: Array [nsppol, nband, ngkpt[0], ngkpt[1], ngkpt[2]] with energies
            in the unit cell mesh in unit `unit`.
        fermie: Fermi energy.
        unit=Unit of input `ucdata_sbk` and `fermie`. Energies will be converted to Hartree before writing.

    .. note::

        #. The k-points must span the reciprocal unit cell, not the Brillouin zone.

        #. The mesh must be closed and centered on Gamma.

        #. Energies are written in row-major (i.e. C) order.

        #  Energies are in Hartree.

    See also http://www.xcrysden.org/doc/XSF.html
    """
    # Xscryden uses Ha for energies.
    ucdata_sbk = EnergyArray(ucdata_sbk, unit).to("Ha")
    fermie = Energy(fermie, unit).to("Ha")

    ucdata_sbk = np.reshape(ucdata_sbk, (nsppol, nband, np.prod(ngkpt)))

    close_it = False
    if not hasattr(file, "write"):
        file = open(str(file), mode="wt")
        close_it = True

    fw = file.write

    # Write the header.
    fw('BEGIN_INFO\n')
    fw('# Band-XCRYSDEN-Structure-File for Visualization of Fermi Surface generated by the ABINIT package\n')
    fw('# NOTE: the first band is relative to spin-up electrons,\n')
    fw('#       the second band to spin-down electrons (if any) and so on ...\n#\n')
    fw('# Launch as: xcrysden --bxsf\n#\n')
    fw(' Fermi Energy: %f\n' % fermie)
    fw('END_INFO\n\n')
    fw('BEGIN_BLOCK_BANDGRID_3D\n')
    fw(' band_energies\n')
    fw(' BEGIN_BANDGRID_3D\n')

    fw(str(nsppol * nband) + "\n")  # Number of bands written.
    fw("%d %d %d\n" % tuple(ngkpt)) # Number of division in the full BZ mesh.
    fw("0 0 0\n")                   # NB: Unshifted meshes are not supported.

    # Reciprocal lattice vectors in Ang^{-1}
    gcell = structure.lattice.reciprocal_lattice.matrix
    for i in range(3):
        fw('%f %f %f\n' % tuple(gcell[i]))

    # Write energies on the full mesh for all spins and bands.
    idx = 0
    for band in range(nband):
        for spin in range(nsppol):
            idx += 1
            enes = ucdata_sbk[spin, band, :]
            fw(" BAND: %d\n" % idx)
            fw("\n".join("%.18e" % v for v in enes))
            fw("\n")

    fw(' END_BANDGRID_3D\n')
    fw('END_BLOCK_BANDGRID_3D\n')
    file.flush()

    if close_it:
        file.close()



#!/usr/bin/env python
r"""
Phonon bands with LO-TO from PHBST.nc
=====================================

This example shows how to plot the phonon band structure of AlAs
including the LO-TO splitting.
These resultas tutorial/lesson_rf2.html
"""

#%%
# Open PHBST file produced by anaddb and extract the phonon bands object.
# (alternatively one can use the shell and `abiopen.py OUT_PHBST.nc -nb`
# to open the file in a jupyter notebook.

from abipy.abilab import abiopen
import abipy.data as abidata

with abiopen(abidata.ref_file("ZnSe_hex_886.out_PHBST.nc")) as ncfile:
    phbands = ncfile.phbands

#%%
# Phonon frequencies with non analytical contributions, if calculated, are saved
# in the anaddb.nc file produced by anaddb. The results should be fetched from there
# and added to the phonon bands.
# Notice that all the directions starting from or arriving at gamma that are used
# in the path should explicitely calculated, even if the values are the same.

phbands.read_non_anal_from_file(abidata.ref_file("ZnSe_hex_886.anaddb.nc"))

#%%
# Plot the phonon frequencies. Note that the labels for the q-points
# are found automatically by searching in an internal database.

phbands.plot(title="ZnSe with LO-TO splitting")

#%%
# For the plotly version, use:

phbands.plotly(title="ZnSe with LO-TO splitting")



#!/usr/bin/env python
r"""
Band structure flow with factory functions
==========================================

This example show how to build a flow to compute the band structure and the DOS of silicon.
Input files are automatically generated with factory functions designed for automatic calculations.
"""
import sys
import os
import abipy.data as abidata
import abipy.flowtk as flowtk

from abipy import abilab


def build_flow(options):
    # Working directory (default is the name of the script with '.py' removed and "run_" replaced by "flow_")
    if not options.workdir:
        options.workdir = os.path.basename(sys.argv[0]).replace(".py", "").replace("run_", "flow_")

    # Initialize structure and pseudos.
    structure = abilab.Structure.from_file(abidata.cif_file("si.cif"))
    pseudos = abidata.pseudos("14si.pspnc")

    # Initialize the flow.
    flow = flowtk.Flow(workdir=options.workdir, manager=options.manager)

    # Use the ebands_input factory function to build a MultiDataset.
    # keyword args are optional (default values are given or computed automatically, see docs).
    multi = abilab.ebands_input(structure, pseudos, kppa=40, dos_kppa=80,
                                nscf_nband=6, ndivsm=10, ecut=6, spin_mode="unpolarized")

    work = flowtk.BandStructureWork(scf_input=multi[0], nscf_input=multi[1], dos_inputs=multi[2])
    flow.register_work(work)

    return flow


# This block generates the thumbnails in the AbiPy gallery.
# You can safely REMOVE this part if you are using this script for production runs.
if os.getenv("READTHEDOCS", False):
    __name__ = None
    import tempfile
    options = flowtk.build_flow_main_parser().parse_args(["-w", tempfile.mkdtemp()])
    build_flow(options).graphviz_imshow()


@flowtk.flow_main
def main(options):
    """
    This is our main function that will be invoked by the script.
    flow_main is a decorator implementing the command line interface.
    Command line args are stored in `options`.
    """
    return build_flow(options)


if __name__ == "__main__":
    sys.exit(main())

############################################################################
#
# Run the script with:
#
#     run_hs_si_ebands.py -s
#
# Use:
#
#    abirun.py flow_h_si_ebands deps
#
# To print the connections among the tasks.
#
# .. code-block:: bash
#
#     <ScfTask, node_id=241095, workdir=flow_ht_si_ebands/w0/t0>
#
#     <NscfTask, node_id=241096, workdir=flow_ht_si_ebands/w0/t1>
#       +--<ScfTask, node_id=241095, workdir=flow_ht_si_ebands/w0/t0>
#
#     <NscfTask, node_id=241097, workdir=flow_ht_si_ebands/w0/t2>
#       +--<ScfTask, node_id=241095, workdir=flow_ht_si_ebands/w0/t0>
#
# ``w0/t1`` is a band structure calculation along a k-path
#  while ``w0/t2`` produced KS eigenvalues in the IBZ for the DOS.
#
#  You can use ipython to plot the electronic bands with the DOS:
#
# .. code-block:: ipython
#
#    In [1]: %matplotlib
#    In [2]: from abipy import abilab
#    In [3]: gsr_kpath = abilab.abiopen("flow_ht_si_ebands/w0/t1/outdata/out_GSR.nc")
#    In [4]: gsr_kmesh = abilab.abiopen("flow_ht_si_ebands/w0/t2/outdata/out_GSR.nc")
#    In [5]: gsr_kpath.ebands.plot_with_edos(gsr_kmesh.ebands.get_edos())
#
# .. image:: https://github.com/abinit/abipy_assets/blob/master/run_ht_si_ebands.png?raw=true
#    :alt: Band structure of Si in the IBZ and along a k-path
#



"""
Classes to compute vibrational properties with phonopy and ML potentials.
"""
from __future__ import annotations

import numpy as np
import json
import abipy.core.abinit_units as abu

from monty.dev import requires
from monty.string import list_strings, marquee
from monty.termcolor import cprint
from ase.calculators.calculator import Calculator
from ase.atoms import Atoms
from pymatgen.io.phonopy import get_phonopy_structure
from abipy.core.structure import Structure
from abipy.dfpt.ddb import DdbFile
from abipy.tools.context_managers import Timer
from abipy.ml.aseml import RX_MODE, CalcBuilder, AseResults, MlBase, relax_atoms, dataframe_from_results_list
try:
    import phonopy
    from phonopy import Phonopy
    from phonopy.structure.atoms import PhonopyAtoms
except ImportError:
    phonopy = None
    Phonopy = None


def cprint_traceback(color="red") -> None:
    """Print traceback."""
    import traceback
    from monty.termcolor import cprint
    cprint(traceback.format_exc(), color=color)


@requires(phonopy, "phonopy should be installed to calculate phonons")
def get_phonopy(structure: Structure,
                supercell_matrix,
                calculator: Calculator,
                distance=0.01,
                primitive_matrix=None,
                remove_drift=True,
                ) -> Phonopy:
    """
    Args:
        structure: Structure object.
        supercell_matrix: Supercell matrix.
        calculator: ASE calculator to be attached to the atoms.
        distance: Distance of finite displacements in Angstrom.
        primitive_matrix
        remove_drift: True if the drift in the forces should be removed.

    Based on a similar implementation available at: https://github.com/modl-uclouvain/randomcarbon/blob/main/randomcarbon/run/phonon.py
    """
    unitcell = Structure.as_structure(structure).get_phonopy_atoms()

    phonon = Phonopy(unitcell, supercell_matrix=supercell_matrix, primitive_matrix=primitive_matrix)
    phonon.generate_displacements(distance=distance)

    forces_list = []
    nsc = len(phonon.supercells_with_displacements)
    print(f"Calculating forces for {nsc} supercell configurations ...")

    for i, sc in enumerate(phonon.supercells_with_displacements):
        a = Atoms(symbols=sc.symbols,
                  positions=sc.positions,
                  masses=sc.masses,
                  cell=sc.cell,
                  pbc=True,
                  calculator=calculator)


        forces = a.get_forces()
        if remove_drift:
            drift_force = forces.sum(axis=0)
            for force in forces:
                force -= drift_force / forces.shape[0]

        forces_list.append(forces)
        print(f"\t{i+1} of {nsc} completed ...")

    phonon.produce_force_constants(forces_list)

    return phonon


class MlPhonopyWithDDB(MlBase):
    """
    Compute phonons with phonopy and a ML potential starting from a DDB file and compare the results.
    """
    def __init__(self, ddb_filepath,
                 distance, asr, dipdip, line_density, qppa,
                 relax_mode, fmax, pressure, steps, optimizer, nn_names,
                 verbose, workdir, prefix=None,
                 supercell=None):
        """
        Args:
            ddb_filepath: DDB filepath.
            distance: Distance of finite displacement in Angstrom.
            asr: Enforce acoustic sum-rule. Abinit variable.
            dipdip: Treatment of dipole-dipole term. Abinit variable.
            line_density: Defines the a density of k-points per reciprocal atom to plot the phonon dispersion.
            qppa: Defines the homogeneous q-mesh used for the DOS in units of q-points per atom.
            relax_mode: "ions" to relax ions only, "cell" for ions + cell, "no" for no relaxation.
            fmax: tolerance for relaxation convergence. Here fmax is a sum of force and stress forces.
            pressure: Target pressure.
            steps: max number of steps for relaxation.
            optimizer: name of the ASE optimizer to use for relaxation.
            nn_names: String or list of strings defining the NN potential. See also CalcBuilder.
            verbose: Verbosity level.
            workdir: Working directory, None to generate temporary directory automatically.
            prefix: Prefix for workdir
            supercell: with supercell dimensions. None to use the supercell from the DDB file.
        """
        super().__init__(workdir, prefix)

        self.distance = float(distance)
        self.asr = asr
        self.dipdip = dipdip
        self.relax_mode = relax_mode
        RX_MODE.validate(self.relax_mode)
        self.fmax = fmax
        self.pressure = pressure
        self.steps = steps
        self.optimizer = optimizer
        self.nn_names = list_strings(nn_names)
        self.verbose = verbose
        self.supercell = supercell
        self.line_density = line_density
        self.qppa = qppa

        self.ddb_filepath = ddb_filepath
        with DdbFile(self.ddb_filepath) as ddb:
            self.initial_atoms = ddb.structure.to_ase_atoms()
            if self.supercell is None:
                # Take supercell from the q-mesh used in the DDB.
                self.supercell = np.eye(3) * ddb.guessed_ngqpt

        if self.supercell is None:
            raise ValueError("Supercell must be specified in input!")

        self.abi_nac_params = {}

    def to_string(self, verbose=0):
        """String representation with verbosity level `verbose`."""
        supercell = self.supercell.tolist()
        s = f"""\

{self.__class__.__name__} parameters:

     ddb_path      = {self.ddb_filepath}
     supercell     = {supercell}
     distance      = {self.distance}
     qppa          = {self.qppa}
     asr           = {self.asr}
     dipdip        = {self.dipdip}
     line_density  = {self.line_density}
     relax_mode    = {self.relax_mode}
     fmax          = {self.fmax}
     steps         = {self.steps}
     optimizer     = {self.optimizer}
     pressure      = {self.pressure}
     nn_names      = {self.nn_names}
     workdir       = {self.workdir}
     verbose       = {self.verbose}
"""
        return s

    def run(self) -> None:
        """
        Run MlPhonopyWithDDB.
        """
        # Run anaddb computation and save results in self.
        with DdbFile(self.ddb_filepath) as ddb, Timer(header="Starting anaddb ph-bands computation...") as timer:
            if ddb.has_lo_to_data and self.dipdip != 0:
                # according to the phonopy website 14.399652 is not the coefficient for abinit
                # probably it relies on the other conventions in the output.
                out = ddb.anaget_epsinf_and_becs()
                self.abi_nac_params = {"born": out.becs.values, "dielectric": out.epsinf, "factor": 14.399652}

            # ab-initio phonons from the DDB.
            with ddb.anaget_phbst_and_phdos_files(
                        qppa=self.qppa, line_density=self.line_density,
                        asr=self.asr, dipdip=self.dipdip, verbose=self.verbose) as g:

                phbst_file, phdos_file = g[0], g[1]
                self.abi_phbands = phbst_file.phbands
                # The q-points passed to phonopy.
                self.py_qpoints = [[qpt.frac_coords for qpt in self.abi_phbands.qpoints]]

        data = {}
        data["ddb"] = self.abi_phbands.get_phfreqs_stats_dict()
        d_nn = data["nn_names"] = {}
        for nn_name in self.nn_names:
            try:
                d = self._run_nn_name(nn_name)
                d_nn[nn_name] = d
                d_nn[nn_name]["exception"] = None
            except Exception as exc:
                #raise
                cprint_traceback()
                d_nn[nn_name] = {}
                d_nn[nn_name]["exception"] = str(exc)

        self.write_json("data.json", data, info="JSON file with final results.")
        self._finalize()

    def _run_nn_name(self, nn_name: str) -> None:
        """
        Run calculation for a single NN potential.
        """
        workdir = self.workdir

        calculator = CalcBuilder(nn_name).get_calculator()
        atoms = self.initial_atoms.copy()
        atoms.calc = calculator
        natom = len(atoms)

        if self.relax_mode == RX_MODE.cell:
            raise RuntimeError("One should not relax the cell when comparing ML phonons with a DDB file!")

        if self.relax_mode != RX_MODE.no:
            print(f"Relaxing input DDB atoms with relax mode: {self.relax_mode}.")
            relax_kws = dict(optimizer=self.optimizer,
                             relax_mode=self.relax_mode,
                             fmax=self.fmax,
                             pressure=self.pressure,
                             steps=self.steps,
                             traj_path=self.get_path(f"{nn_name}_relax.traj", "ASE relax trajectory"),
                             verbose=self.verbose,
                            )

            relax = relax_atoms(atoms, **relax_kws)
            relax.summarize(["DDB_initial", "DDB_relaxed"])
            #print(relax)

        # Call phonopy to compute phonons with finite difference and ML potential.
        # Include non-analytical term if dipoles are available in the DDB file.
        with Timer(header=f"Calling get_phonopy with {nn_name=}", footer=""):
            phonon = get_phonopy(atoms, self.supercell, calculator,
                                 distance=self.distance, primitive_matrix=None, remove_drift=True)
            if self.abi_nac_params:
                print("Including dipolar term in phonopy using BECS and eps_inf taken from DDB.")
                phonon.nac_params = self.abi_nac_params

        # Save phonopy object in Yaml format.
        phonon.save(filename=workdir / f"phonopy_params.yaml", settings={'force_constants': True})

        from abipy.dfpt.phonons import PhononBands, PhononBandsPlotter
        with Timer(header="Starting phonopy ph-band computation...", footer=""):
            phonon.run_band_structure(self.py_qpoints, with_eigenvectors=True)

        plt = phonon.plot_band_structure()
        plt.savefig(workdir / f"phonopy_{nn_name}_phbands{self.fig_ext}")
        plt.close()

        bands_dict = phonon.get_band_structure_dict()
        nqpt = 0
        py_phfreqs, py_displ_cart = [], []
        for q_list, w_list, eig_list in zip(bands_dict['qpoints'], bands_dict['frequencies'], bands_dict['eigenvectors']):
            nqpt += len(q_list)
            py_phfreqs.extend(w_list)
            py_displ_cart.extend(eig_list)

        py_phfreqs = np.reshape(py_phfreqs, (nqpt, 3*natom)) / abu.eV_to_THz
        py_displ_cart = np.reshape(py_displ_cart, (nqpt, 3*natom, 3*natom))

        # Build abipy phonon bands from phonopy results.
        py_phbands = PhononBands(self.abi_phbands.structure, self.abi_phbands.qpoints, py_phfreqs,
                                 # FIXME: Use phononopy displacement
                                 self.abi_phbands.phdispl_cart,
                                 non_anal_ph=None,
                                 amu=self.abi_phbands.amu,
                                 epsinf=self.abi_phbands.epsinf,
                                 zcart=self.abi_phbands.zcart,
                                 )

        # Compute diff stats.
        mabs_wdiff_ev = np.abs(py_phbands.phfreqs - self.abi_phbands.phfreqs).mean()

        ph_plotter = PhononBandsPlotter(key_phbands=[
            (f"phonopy with {nn_name}", py_phbands),
            ("ABINIT DDB", self.abi_phbands),
        ])
        mae_str = f"MAE {1000 * mabs_wdiff_ev:.3f} meV"
        print(mae_str)
        latex_formula = self.abi_phbands.structure.latex_formula
        ph_plotter.combiplot(show=False, title=f"{latex_formula}: {mae_str}", units="meV",
                             savefig=str(workdir / f"combiplot_{nn_name}.pdf"))

        data = dict(
            mabs_wdiff_ev=mabs_wdiff_ev,
            **py_phbands.get_phfreqs_stats_dict()
        )

        # Compute phonon DOS and generate file with figure.
        #phonon.auto_total_dos(plot=True)
        #plt.savefig(workdir / f"phonopy_{nn_name}_phdos{self.fig_ext}")
        #plt.close()

        # Compute site-project phonon DOS and generate file with figure.
        #phonon.auto_projected_dos(plot=True)
        #plt.savefig(workdir / f"phonopy_{nn_name}_pjdos{self.fig_ext}")
        #plt.close()

        #phonon.run_thermal_properties(t_step=10, t_max=1000, t_min=0)
        #phonon.write_yaml_thermal_properties(filename=workdir / f"phonopy_{nn_name}_thermal_properties.yaml")

        return data


class MlPhonopy(MlBase):
    """
    Compute phonons with phonopy and ML potential.
    """
    def __init__(self, structure, supercell, distance, line_density, qppa,
                 relax_mode, fmax, pressure, steps, optimizer, nn_names,
                 verbose, workdir, prefix=None):
        """
        Args:
            structure: Structure object
            supercell: Supercell dimensions.
            distance: Distance of finite displacements in Angstrom.
            line_density: Defines the a density of k-points per reciprocal atom to plot the phonon dispersion.
            qppa: Defines the homogeneous q-mesh used for the DOS in units of q-points per atom.
            relax_mode: "ions" to relax ions only, "cell" for ions + cell, "no" for no relaxation.
            fmax: tolerance for relaxation convergence. Here fmax is a sum of force and stress forces.
            pressure: Target pressure.
            steps: max number of steps for relaxation.
            optimizer: name of the ASE optimizer to use for relaxation.
            nn_names: String or list of strings defining the NN potential. See also CalcBuilder.
            verbose: Verbosity level.
            workdir: Working directory, None to generate temporary directory automatically.
            prefix: Prefix for workdir.
        """
        super().__init__(workdir, prefix)

        self.initial_atoms = structure.to_ase_atoms()

        self.distance = float(distance)
        #self.asr = asr
        #self.dipdip = dipdip
        self.relax_mode = relax_mode
        RX_MODE.validate(self.relax_mode)
        self.fmax = fmax
        self.pressure = pressure
        self.steps = steps
        self.optimizer = optimizer
        self.nn_names = list_strings(nn_names)
        self.verbose = verbose
        self.supercell = supercell
        self.line_density = line_density
        self.qppa = qppa
        #self.abi_nac_params = {}

    def to_string(self, verbose=0):
        """String representation with verbosity level `verbose`."""
        supercell = self.supercell.tolist()
        s = f"""\

{self.__class__.__name__} parameters:

     supercell     = {supercell}
     distance      = {self.distance}
     qppa          = {self.qppa}
     line_density  = {self.line_density}
     relax_mode    = {self.relax_mode}
     fmax          = {self.fmax}
     steps         = {self.steps}
     optimizer     = {self.optimizer}
     pressure      = {self.pressure}
     nn_names      = {self.nn_names}
     workdir       = {self.workdir}
     verbose       = {self.verbose}
"""
        return s

    def run(self) -> None:
        """Run calculation."""
        data = {}
        d_nn = data["nn_names"] = {}
        for nn_name in self.nn_names:
            try:
                d = self._run_nn_name(nn_name)
                d_nn[nn_name] = d
                d_nn[nn_name]["exception"] = None
            except Exception as exc:
                cprint_traceback()
                d_nn[nn_name] = {}
                d_nn[nn_name]["exception"] = str(exc)

        self.write_json("data.json", data, info="JSON file with final results.")
        self._finalize()

    def _run_nn_name(self, nn_name: str) -> None:
        """
        Run calculation for a single NN potential.
        """
        workdir = self.workdir

        calculator = CalcBuilder(nn_name).get_calculator()
        atoms = self.initial_atoms.copy()
        atoms.calc = calculator
        natom = len(atoms)

        if self.relax_mode != RX_MODE.no:
            print(f"Relaxing input atoms with relax mode: {self.relax_mode}.")
            relax_kws = dict(optimizer=self.optimizer,
                             relax_mode=self.relax_mode,
                             fmax=self.fmax,
                             pressure=self.pressure,
                             steps=self.steps,
                             traj_path=self.get_path(f"{nn_name}_relax.traj", "ASE relax trajectory"),
                             verbose=self.verbose,
                            )

            relax = relax_atoms(atoms, **relax_kws)
            relax.summarize(["initial", "relaxed"])
            print(relax)

        # Call phonopy to compute phonons with finite difference and ML potential.
        # Include non-analytical term if dipoles are available in the DDB file.
        with Timer(header=f"Calling get_phonopy with {nn_name=}", footer=""):
            phonon = get_phonopy(atoms, self.supercell, calculator,
                                 distance=self.distance, primitive_matrix=None, remove_drift=True)

        plt = phonon.auto_band_structure(
              npoints=101,
              with_eigenvectors=False,
              with_group_velocities=False,
              plot=True,
              write_yaml=True,
              filename=workdir / f"{nn_name}_band.yml",
        )
        plt.savefig(workdir / f"phonopy_{nn_name}_phbands{self.fig_ext}")
        plt.close()

        # Save phonopy object in Yaml format.
        phonon.save(filename=workdir / f"phonopy_params.yaml", settings={'force_constants': True})

        # Compute phonon DOS and generate file with figure.
        phonon.auto_total_dos(plot=True)
        plt.savefig(workdir / f"phonopy_{nn_name}_phdos{self.fig_ext}")
        plt.close()

        # Compute site-project phonon DOS and generate file with figure.
        #phonon.auto_projected_dos(plot=True)
        #plt.savefig(workdir / f"phonopy_{nn_name}_pjdos{self.fig_ext}")
        #plt.close()

        # Compute thermal properties.
        phonon.run_thermal_properties(t_step=10, t_max=1000, t_min=0)
        phonon.write_yaml_thermal_properties(filename=workdir / f"phonopy_{nn_name}_thermal_properties.yaml")

        phonon.plot_thermal_properties()
        plt.savefig(workdir / f"phonopy_{nn_name}_thermal_properties{self.fig_ext}")
        plt.close()

        return {}




#!/usr/bin/env python
r"""
Eliashberg function
===================

This example shows how to plot the Eliashberg function a2F(w)
and the total e-ph coupling strenght in metals.
"""
from abipy import abilab
import abipy.data as abidata

a2file = abilab.abiopen(abidata.ref_file("al_888k_161616q_A2F.nc"))

#print(a2file)
#a2file.phbands.plot()
#a2file.a2f_qintp.plot()
#with_lambda = False
#fig = a2file.a2f_qcoarse.plot_nuterms(with_lambda=with_lambda, show=False)
#a2file.a2f_qintp.plot_nuterms(axmat=fig.axes, with_lambda=with_lambda)

#a2file.plot()

phdos_path = abidata.ref_file("al_161616q_PHDOS.nc")
a2file.plot_with_a2f(phdos=phdos_path)

a2file.plot_eph_strength(what="gamma")
#a2file.plot_eph_strength(what="lambda")

a2file.plot_a2f_interpol()

# Grid with 3 plots (a2F, F, a2F) with F taken from PHDOS file.
#a2file.a2f_qintp.plot_a2(phdos_path)



# coding: utf-8
"""
Object to plot DFPT potentials in the phonon mode representation.
"""
import numpy as np

#from collections import OrderedDict
from monty.string import marquee
from monty.functools import lazy_property
#from abipy.tools.plotting import add_fig_kwargs, get_ax_fig_plt
from abipy.core.mixins import AbinitNcFile, Has_Structure, NotebookWriter
#from abipy.core.kpoints import Kpoint #KpointList,
#from abipy.tools import duck
from abipy.iotools import xsf, ETSF_Reader #, cube Visualizer,


class V1qnuFile(AbinitNcFile, Has_Structure, NotebookWriter):

    def __init__(self, filepath):
        super().__init__(filepath)
        self.reader = r = ETSF_Reader(filepath)
        # Read dimensions.
        self.nfft = r.read_dimvalue("nfft")
        self.nspden = r.read_dimvalue("nspden")
        self.natom3 = len(self.structure) * 3
        # Read FFT grid.
        self.ngfft = r.read_value("ngfft")

    @lazy_property
    def structure(self):
        """|Structure| object."""
        return self.reader.read_structure()

    def close(self):
        self.reader.close()

    @lazy_property
    def params(self):
        """:class:`OrderedDict` with parameters that might be subject to convergence studies."""
        return {}

    def __str__(self):
        return self.to_string()

    def to_string(self, verbose=0):
        """String representation."""
        lines = []; app = lines.append
        app(marquee("File Info", mark="="))
        app(self.filestat(as_string=True))
        app("")
        app(self.structure.to_string(verbose=verbose, title="Structure"))
        app("")
        for iq, qpt in enumerate(self.qpoints):
            app("[%d] %s" % (iq, repr(qpt)))

        return "\n".join(lines)

    #@lazy_property
    #def qpoints(self):
    #    return KpointList(self.structure.reciprocal_lattice, frac_coords=self.reader.read_value("qlist"))

    #def _find_iqpt_qpoint(self, qpoint):
    #    if duck.is_intlike(qpoint):
    #        iq = qpoint
    #        qpoint = self.qpoints[iq]
    #    else:
    #        qpoint = Kpoint.as_kpoint(qpoint, self.structure.reciprocal_lattice)
    #        iq = self.qpoints.index(qpoint)

    #    return iq, qpoint

    def visualize_nu(self, nu, spin=0, appname="vesta"):
        #iq, qpoint = self._find_iqpt_qpoint(qpoint)

        def xsf_write(filename, datar):
            with open(filename, mode="wt") as fh:
                xsf.xsf_write_structure(fh, self.structure)
                xsf.xsf_write_data(fh, self.structure, datar, add_replicas=True)

        from abipy.tools.numtools import transpose_last3dims
        wqnu = self.reader.read_variable("phfreqs")[nu]

        for vname in ("v1_qnu", "v1lr_qnu"):
            # Fortran array nctkarr_t("v1_qnu", "dp", "two, nfft, nspden, natom3")])
            datar = self.reader.read_variable(vname)[nu, spin]
            datar = datar[:, 0] + 1j * datar[:, 1]
            #datar /= np.sqrt(2 * wqnu)
            datar = transpose_last3dims(datar)
            datar = np.reshape(np.abs(datar), self.ngfft)
            xsf_write("%s.xsf" % vname, datar)

        #visu = Visualizer.from_name(appname)
        #ext = "xsf"
        #if ext not in visu.supported_extensions():
        #    raise ValueError("Visualizer %s does not support XSF files" % visu)
        #from abipy.core.globals import abinb_mkstemp
        #_, filename = abinb_mkstemp(suffix="." + ext, text=True)

    #@add_fig_kwargs
    #def plot_v1qnu_vs_lr(self, ax=None, fontsize=8, **kwargs):
    #    """
    #    Plot the difference between the ab-initio v1_qnu and the potential obtained with Verdi's model

    #    Args:
    #        ax: |matplotlib-Axes| or None if a new figure should be created.
    #        fontsize: Label and title fontsize.

    #    Return: |matplotlib-Figure|
    #    """
    #    # Fortran array nctkarr_t("v1_qnu", "dp", "two, nfft, nspden, natom3, nqlist")])
    #    v1_qnu = self.reader.read_value("v1_qnu", cmode="c")
    #    v1lr_qnu = self.reader.read_value("v1lr_qnu", cmode="c")

    #    stats = OrderedDict([
    #        ("min", []),
    #        ("max", []),
    #        ("mean", []),
    #        ("std", []),
    #    ])

    #    qnorms = []
    #    for iq, qpt in enumerate(self.qpoints):
    #        qnorms.append(qpt.norm)
    #        abs_diff = np.abs(v1_qnu[iq] - v1lr_qnu[iq])
    #        for key in stats.keys():
    #            stats[key].append(getattr(abs_diff, key)())

    #    # Sort values by |q|.
    #    qindex = np.arange(len(qnorms))
    #    items = sorted(list(zip(qnorms, qindex)),  key=lambda t: t[0])
    #    qnorm = np.array([t[0] for t in items])
    #    qindex = np.array([t[1] for t in items])
    #    #print(qnorm, "\n", qindex)
    #    for key in stats.keys():
    #        stats[key] = np.array(stats[key])[qindex]

    #    ax, fig, plt = get_ax_fig_plt(ax=ax)
    #    for key, values in stats.items():
    #        ax.plot(values, label=key)

    #    ax.grid(True)
    #    ax.set_xlabel(r"$|\bf{q}|$ 1/Ang")
    #    #ax.set_ylabel(r"$|g_{\bf q}|$ (meV)")
    #    ax.legend(loc="best", fontsize=fontsize, shadow=True)
    #    #title = "band_kq: %s, band_k: %s, kpoint: %s" % (band_kq, band_k, repr(kpoint))
    #    #ax.set_title(title, fontsize=fontsize)
    #    return fig

    def yield_figs(self, **kwargs):  # pragma: no cover
        """
        This function *generates* a predefined list of matplotlib figures with minimal input from the user.
        """
        yield self.plot_v1qnu_vs_lr(show=False)

    def write_notebook(self, nbpath=None):
        """
        Write a jupyter_ notebook to ``nbpath``. If nbpath is None, a temporay file in the current
        working directory is created. Return path to the notebook.
        """
        nbformat, nbv, nb = self.get_nbformat_nbv_nb(title=None)

        nb.cells.extend([
            nbv.new_code_cell("ncfile = abilab.abiopen('%s')" % self.filepath),
            nbv.new_code_cell("print(ncfile)"),
            nbv.new_code_cell("# ncfile.visualize_qpoint_nu(qpoint, nu, spin=0, appname='vesta'")
        ])

        return self._write_nb_nbpath(nb, nbpath)


if __name__ == "__main__":
    import sys
    with V1qnuFile(sys.argv[1]) as ncfile:
        ncfile.visualize_nu(nu=5)



# coding: utf-8
"""
Collection of low-level tools to facilitate the interface with resource managers.

The preferred way of importing this module is:

    import qutils as qu
"""
from __future__ import annotations

import os

from subprocess import Popen, PIPE, run
from monty.string import is_string
from pymatgen.core.units import Time, Memory, UnitError
from abipy.tools.typing import PathLike
from abipy.tools import duck
from abipy.tools.text import rm_multiple_spaces


def slurm_parse_timestr(s: str) -> Time:
    """
    A slurm time parser. Accepts a string in one the following forms:

        # "days-hours",
        # "days-hours:minutes",
        # "days-hours:minutes:seconds".
        # "minutes",
        # "minutes:seconds",
        # "hours:minutes:seconds",

    Returns: Time in seconds.

    Raises:
        `ValueError` if string is not valid.
    """
    days, hours, minutes, seconds = 0, 0, 0, 0

    if duck.is_number_like(s):
        return Time(s, "s")

    if '-' in s:
        # "days-hours",
        # "days-hours:minutes",
        # "days-hours:minutes:seconds".
        days, s = s.split("-")
        days = int(days)

        if ':' not in s:
            hours = int(float(s))
        elif s.count(':') == 1:
            hours, minutes = map(int, s.split(':'))
        elif s.count(':') == 2:
            hours, minutes, seconds = map(int, s.split(':'))
        else:
            raise ValueError("More that 2 ':' in string!")

    else:
        # "minutes",
        # "minutes:seconds",
        # "hours:minutes:seconds",
        if ':' not in s:
            minutes = int(float(s))
        elif s.count(':') == 1:
            minutes, seconds = map(int, s.split(':'))
        elif s.count(':') == 2:
            hours, minutes, seconds = map(int, s.split(':'))
        else:
            raise ValueError("More than 2 ':' in string!")

    return Time((days*24 + hours)*3600 + minutes*60 + seconds, "s")


def time2slurm(timeval: float, unit="s") -> str:
    """
    Convert a number representing a time value in the given unit (Default: seconds)
    to a string following the slurm convention: "days-hours:minutes:seconds".

    >>> assert time2slurm(61) == '0-0:1:1' and time2slurm(60*60+1) == '0-1:0:1'
    >>> assert time2slurm(0.5, unit="h") == '0-0:30:0'
    """
    d, h, m, s = 24*3600, 3600, 60, 1

    timeval = Time(timeval, unit).to("s")
    days, hours = divmod(timeval, d)
    hours, minutes = divmod(hours, h)
    minutes, secs = divmod(minutes, m)

    return "%d-%d:%d:%d" % (days, hours, minutes, secs)


def time2pbspro(timeval: float, unit="s") -> str:
    """
    Convert a number representing a time value in the given unit (Default: seconds)
    to a string following the PbsPro convention: "hours:minutes:seconds".

    >>> assert time2pbspro(2, unit="d") == '48:0:0'
    """
    h, m, s = 3600, 60, 1

    timeval = Time(timeval, unit).to("s")
    hours, minutes = divmod(timeval, h)
    minutes, secs = divmod(minutes, m)

    return "%d:%d:%d" % (hours, minutes, secs)


def time2loadlever(timeval: float, unit="s") -> str:
    """
    Convert a number representing a time value in the given unit (Default: seconds)
    to a string following the LoadLever convention. format hh:mm:ss (hours:minutes:seconds)

    >>> assert time2loadlever(2, unit="d") == '48:00:00'
    """
    h, m, s = 3600, 60, 1

    timeval = Time(timeval, unit).to("s")
    hours, minutes = divmod(timeval, h)
    minutes, secs = divmod(minutes, m)

    return "%d:%02d:%02d" % (hours, minutes, secs)


def timelimit_parser(s):
    """Convert a float or a string into time in seconds."""
    try:
        return Time(float(s), "s")
    except ValueError:
        return slurm_parse_timestr(s)


def any2mb(s):
    """Convert string or number to memory in megabytes."""
    if is_string(s):
        try:
            # latest pymatgen version (as of july 2024)
            mem = int(Memory.from_str(s.upper()).to("MB"))
        except (KeyError, UnitError):  # For backward compatibility with older pymatgen versions
            try:
                mem = int(Memory.from_str(s.replace("B", "b")).to("Mb"))
            except AttributeError:  # For even older pymatgen versions
                mem = int(Memory.from_string(s.replace("B", "b")).to("Mb"))
        return mem
    else:
        return int(s)


def slurm_get_jobs() -> dict[int, dict]:
    """
    Invoke squeue, parse output and return list of dictionaries with job info indexed by job id.
    """
    # Based on https://gist.github.com/stevekm/7831fac98473ea17d781330baa0dd7aa
    process = Popen(["squeue", "--me", "-o", '%all'],
                       stdout=PIPE, stderr=PIPE, shell=False, universal_newlines=True)
    proc_stdout, proc_stderr = process.communicate()

    lines = proc_stdout.split('\n')
    header_line = lines.pop(0)
    header_cols = header_line.split('|')
    entries = []
    error_lines = [] # do something with this later
    for line in lines:
        parts = line.split('|')
        if len(parts) != len(header_cols):
            error_lines.append((len(parts), line, parts))
        else:
            d = {}
            for i, key in enumerate(header_cols):
                d[key] = parts[i]
                if key == "JOBID":
                    d[key] = int(d[key])
            entries.append(d)

    return {e["JOBID"]: e for e in entries}


class SlurmJobArray:
    """

    Example:

    header = '''\
#!/bin/bash

#SBATCH --account=battab
#SBATCH --job-name=abiml_md
#SBATCH --time=0-16:0:0
#SBATCH --partition=batch
#SBATCH --nodes=1             # 1 node has 128 cores
#SBATCH --ntasks-per-node=1
#SBATCH --cpus-per-task=1

conda activate env3.10
export OMP_NUM_THREADS=$SLURM_CPUS_PER_TASK
ulimit -s unlimited
'''

command = "abiml.py md"
arr_options = ["--help", "--version"]
job_array = SlurmJobArray(header, command, arr_options)
print(job_array)
queue_id = job_array.sbatch("job_array.sh")
    """

    def __init__(self, header: str, command: str, arr_options: list[str]):
        self.command = command
        if not self.command.endswith(" "): self.command += " "
        self.header = header
        self.arr_options = arr_options
        self.arr_options_str = rm_multiple_spaces("\n".join(arr_options))

    def __str__(self):
        # Add slurm array section.
        lines = self.header.splitlines()
        for il, line in enumerate(lines):
            if line.startswith("#SBATCH"):
                break
        else:
            raise ValueError("Cannot find line starting with #SBATCH")

        lines.insert(il, f"#SBATCH --array=0-{len(self.arr_options)-1}")
        header = "\n".join(lines)

        select_opts = r"""
# multiline string
OPTS_STRING="
%s
"

# Convert the multiline string to an array
IFS=$'\n' read -rd '' -a OPTS_LIST <<< "$OPTS_STRING"

# Index of the entry you want (0-based)
index=${SLURM_ARRAY_TASK_ID}

# Check if the index is within the range of the array
OPTS="${OPTS_LIST[index]}"
echo "Running entry at index $index:\nwith OPTS=$OPTS"

env
""" % (self.arr_options_str)

        end = f"""
{self.command} ${{OPTS}} > job_${{index}}.log 2> job_${{index}}.err

# Remove the file with the Slurm job id
me=$(basename "$0")
rm ${{me}}.qid
"""

        return header + select_opts + end

    def sbatch(self, slurm_filepath: PathLike) -> int:
        """
        Write slurm submission script to slurm_filepath and submits it.
        Return Slurm JOB id.
        """
        # Make sure no slurm job is already running by checking for a .qid file.
        path_qid = slurm_filepath + ".qid"
        if os.path.exists(path_qid):
            with open(path_qid, "rt") as fh:
                queue_id = int(fh.read().split("#"))
                err_msg = f"Found slurm job ID {queue_id} in {path_qid}" + \
                          "This usually indicates that a similar array job is already running\n" + \
                          f"If this not the case, please remove {path_qid} and rerun the script."
                raise RuntimeError(err_msg)

        with open(slurm_filepath, "wt") as fh:
            fh.write(str(self))

        queue_id = slurm_sbatch(slurm_filepath)
        return queue_id


def slurm_write_and_sbatch(script_filepath: str, slurm_script_str: str) -> int:
    """
    Write job script and submit it to the queue with Slurm sbatch. Return Slurm JOB ID.
    """
    with open(script_filepath, "wt") as fh:
        fh.write(slurm_script_str)
        return slurm_sbatch(script_filepath)


def slurm_sbatch(slurm_filepath: PathLike) -> int:
    """
    Submit a job script to the queue with Slurm sbatch. Return Slurm JOB ID.
    """
    from monty.os import cd
    dirpath = os.path.dirname(slurm_filepath)
    #print("dirpath", dirpath)
    with cd(dirpath):

        # need string not bytes so must use universal_newlines
        slurm_filepath = str(slurm_filepath)
        process = Popen(['sbatch', slurm_filepath], stdout=PIPE, stderr=PIPE, universal_newlines=True)
        out, err = process.communicate()

        # grab the returncode. SLURM returns 0 if the job was successful
        if process.returncode == 0:
            try:
                # output should of the form '2561553.sdb' or '352353.jessup' - just grab the first part for job id
                queue_id = int(out.split()[3])
                path_qid = slurm_filepath + ".qid"
                print(f"Job submission was successful and queue_id: {queue_id}")
                print("Saving slurm job ID in:", path_qid)
                with open(path_qid, "wt") as fh:
                    fh.write(str(queue_id) + " # Slurm job id")
                return queue_id

            except Exception as exc:
                # probably error parsing job code
                print('Could not parse job id following slurm...')
                raise exc
        else:
            raise RuntimeError(f"Error while submitting {slurm_filepath=} with {process.returncode=},\n{out=}\n{err=}")


def get_sacct_info():
    """
    Run the sacct command to get the job information
    """
    try:

        result = run(['sacct', '--format=JobID,JobName,Partition,Account,AllocCPUS,State,ExitCode', '--noheader'],
                      stdout=PIPE, stderr=PIPE, text=True)

        # Check if the command was successful
        if result.returncode != 0:
            print(f"Error running sacct: {result.stderr}")
            return None

        # Process the output
        jobs_info = result.stdout.strip().split('\n')
        jobs = [dict(zip(['JobID', 'JobName', 'Partition', 'Account', 'AllocCPUS', 'State', 'ExitCode'], job.split())) 
                for job in jobs_info]
        return jobs

    except Exception as e:
        print(f"An error occurred: {e}")
        return None


def get_completed_job_info(job_id: int | str):
    try:
        # Define the fields we want to retrieve
        fields = "JobID,JobName,Partition,Account,AllocCPUS,State,ExitCode,Start,End,Elapsed,TotalCPU,MaxRSS"

        # Run the sacct command with the specified fields for the given job ID
        result = run(
            ['sacct', '--jobs', str(job_id), '--format', fields, '--noheader', '--parsable2'],
            stdout=PIPE, stderr=PIPE, text=True
        )

        # Check if the command was successful
        if result.returncode != 0:
            print(f"Error running sacct: {result.stderr}")
            return None

        # Process the output
        lines = result.stdout.strip().split('\n')
        jobs = [dict(zip(fields.split(','), line.split('|'))) for line in lines]
        return jobs

    except Exception as e:
        print(f"An error occurred: {e}")
        return None


def get_slurm_template(body: str) -> str:
    """
    Return template for slurm submission that is supposed to be customized by the user.
    """
    return f"""\
#!/bin/bash
#
# Please CUSTOMIZE this section according to your cluster and the type of calculation
#
#SBATCH --job-name=my_job
#SBATCH --output=%j_%x.slurm.out
#SBATCH --error=%j_%x.slurm.err
#SBATCH --partition=debug
#SBATCH --nodes=1
#SBATCH --ntasks-per-node=64
#SBATCH --cpus-per-task=1
#SBATCH --mem-per-cpu=2G
#SBATCH --time=2:00:00
#SBATCH --account=htforft

export OMP_NUM_THREADS=$SLURM_CPUS_PER_TASK

# ------------------------------------------------------------------------------
# Printing some information
# ------------------------------------------------------------------------------

echo "------------------- Job info -------------------"
echo "job_id             : $SLURM_JOB_ID"
echo "jobname            : $SLURM_JOB_NAME"
echo "queue              : $SLURM_JOB_PARTITION"
echo "qos                : $SLURM_JOB_QOS"
echo "account            : $SLURM_JOB_ACCOUNT"
echo "submit dir         : $SLURM_SUBMIT_DIR"
echo "number of mpi tasks: $SLURM_NTASKS tasks"
echo "OMP_NUM_THREADS    : $OMP_NUM_THREADS"
echo "number of gpus     : $SLURM_GPUS_ON_NODE"
echo "------------------- Node list ------------------"
echo $SLURM_JOB_NODELIST

echo "---------------- Printing limits ---------------"
ulimit -s unlimited
ulimit -a

# ------------------------------------------------------------------------------
# Setting up the environment
# ------------------------------------------------------------------------------

echo "----------------- Environment ------------------"

source $HOME/vasp.6.2.1/modules.sh
module list

echo "--------------- Running the code ---------------"
echo -n "This run started on: "
date

{body}

echo -n "This run completed on: "
date
"""


def get_custodian_template() -> str:
    return """\
#!/usr/bin/env python

from custodian.custodian import Custodian
from custodian.vasp.jobs import VaspJob
from custodian.vasp.handlers import VaspErrorHandler, UnconvergedErrorHandler

# List of error handlers
handlers = [
    VaspErrorHandler(),        # Handles common VASP errors
    UnconvergedErrorHandler()  # Handles unconverged calculations
]

# Define the VASP job with appropriate command and parameters
jobs = [
    VaspJob(
        #["mpirun", "vasp_std"],  # Replace NCORES with the number of cores
        #["mpirun", "-np", "1", "vasp_std"],  # Replace NCORES with the number of cores
        ["mpirun", "vasp_std"],  # Replace NCORES with the number of cores
        auto_npar=False,
        final=True
    )
]

# Create the Custodian instance with handlers and jobs
c = Custodian(handlers, jobs, max_errors=5)

# Run the Custodian job
c.run()
"""


#!/usr/bin/env python
r"""
Alas mobility
=============

This example shows how to plot the electronic band structure,
phonon dispersion, electron linewidths and
the convergence of the electron mobility in the case of AlAs
(see example e-ph mobility flow).
Only electrons are considered, but the script can easily be
extended to holes (and combination of e/h as well).
"""
#%%

import numpy as np
import matplotlib as mlp
import matplotlib.pyplot as plt
from abipy.abilab import abiopen
import glob
from abipy.eph.rta import RtaRobot

root = "../../flows/flow_eph_mob/"

fz = 13 # Fontsize for text, labels,...
mz = 2  # Markersize for dots (linewidths plot)

fig, axes = plt.subplots(2, 2)
plt.subplots_adjust(wspace=0.35, hspace=0.33)

axes[0, 0].text(-0.1,1.05, '(a)', size=fz, transform=axes[0, 0].transAxes)
axes[1, 0].text(-0.1,1.05, '(b)', size=fz, transform=axes[1, 0].transAxes)
axes[0, 1].text(-0.1,1.05, '(c)', size=fz, transform=axes[0, 1].transAxes)
axes[1, 1].text(-0.1,1.05, '(d)', size=fz, transform=axes[1, 1].transAxes)

###
### Upper left : electronic band structure
###

# Open the GSR file containing the BS
with abiopen(root + "w0/t1/outdata/out_GSR.nc") as abifile:
    ebands = abifile.ebands
    formula = abifile.structure.formula

# Add the title to the whole figure containing the pretty formula
pretty_formula = "".join([i for i in formula if i != str(1) and i != " "]) # Remove '1's and spaces
fig.suptitle(pretty_formula)

# Plot the bands (keep only the grid on the x-axis)
ebands.plot(ax=axes[0,0], show=False, linewidth=1.5, ylims=(-3, 6), ax_grid=False)

# Set label
axes[0, 0].set_ylabel('Energy (eV)', fontsize=fz)
axes[0, 0].set_xlabel(None)

###
### Lower left : phonon dispersion
###

# Open the DDB file and get phonon dispersion
# Can be used to check the phonon dispersion obtained with the e-ph code
with abiopen(root + "w1/outdata/out_DDB") as abifile:
    phbst, phdos = abifile.anaget_phbst_and_phdos_files(ndivsm=10)

# Plot phonon dispersion (keep only the grid on the x-axis)
phbst.phbands.plot(ax=axes[1,0], show=False, units="mev", linewidth=1.5, ax_grid=False)

# Set label
axes[1, 0].set_ylabel('Frequency (meV)', fontsize=fz)
axes[1, 0].set_xlabel(None)

###
### Upper right : linewidths
###

# We get the linewidths for the densest mesh computed
# We open the SIGEPH file corresponding to this task
with abiopen(root + 'w3/t2/outdata/out_SIGEPH.nc') as abifile:
    # First within SERTA
    qparray = abifile.get_qp_array(mode="ks+lifetimes", rta_type="serta")
    qparray = qparray[np.nonzero(qparray)]
    eigs = qparray.real - np.min(qparray.real) # 0 to CBM
    lws = 2000*qparray.imag # Factor or 2x1000 to get from linewidths to 1/tau in meV

    # Then within MRTA
    qparray_mrta = abifile.get_qp_array(mode="ks+lifetimes", rta_type="mrta")
    qparray_mrta = qparray_mrta[np.nonzero(qparray_mrta)]
    eigs_mrta = qparray_mrta.real - np.min(qparray_mrta.real) # 0 to CBM
    lws_mrta = 2000*qparray_mrta.imag

# Plot 1/tau
axes[0, 1].plot(eigs, lws, 'ob', markersize=mz, label='SERTA')
axes[0, 1].plot(eigs_mrta, lws_mrta, 'xr', markersize=mz, label='MRTA')

axes[0, 1].set_xticks([0, 0.25], ['0', '0.25'])

# Set the axis labels and legend
axes[0, 1].set_xlabel(r'$\varepsilon_{n\mathbf{k}} - \varepsilon_{\mathrm{CBM}}$ (eV)', labelpad=2, fontsize=fz)
axes[0, 1].set_ylabel(r'$\tau_{n\mathbf{k}}^{-1}$ (meV)', fontsize=fz)
axes[0, 1].legend(loc='best', labelcolor='linecolor')


###
### Lower right : convergence of the mobility
###

# First we find all the RTA.nc files
abifiles = glob.glob(root+'w*/t*/outdata/*RTA.nc')

# We create a robot with these files and plot the mobilities
robot = RtaRobot.from_files(abifiles)
robot.plot_mobility_kconv(ax=axes[1, 1], eh=0, bte=["ibte", "mrta", "serta"], show=False, ax_grid=False)

# Tune the plot
axes[1, 1].set_title(None)
axes[1, 1].legend(fontsize=fz, framealpha=0.5)
axes[1, 1].set_xlabel('$N_k \\times$ $N_k \\times$ $N_k$ $\\mathbf{k}$-point grid', fontsize=fz)
axes[1, 1].set_ylabel(r'$\mu_e$'+' (cm$^2$/(V$\\cdot$s))', fontsize=fz)

# Reactivate the grid on the x-axis for the band structures
axes[0, 0].grid(True, axis='x')
axes[1, 0].grid(True, axis='x')

# We save the figure in pdf format
fig.savefig(pretty_formula + ".pdf", bbox_inches='tight')



#!/usr/bin/env python
r"""
Bethe-Salpeter
==============

This example shows how to plot the macroscopic dielectric function (MDF)
computed in the Bethe-Salpeter code.
"""
import abipy.data as abidata
from abipy.abilab import abiopen

# Open the MDF file produced in the tutorial.
mdf_file = abiopen(abidata.ref_file("tbs_4o_DS2_MDF.nc"))

# Plot the imaginary part of the macroscopic
# dielectric function (EXC, RPA, GWRPA) between 2 and 5 eV.
xlims = (2, 5)

mdf_file.plot_mdfs(xlims=xlims,
                   title="Si absorption spectrum: EXC vs RPA (averaged over small q-points)")

# Plot the real part for the first q-point --> 0
mdf_file.plot_mdfs(cplx_mode="Re", qpoint=0, xlims=xlims,
                   title="Real part for the first q-point: EXC vs RPA")

# Plot the 6 different components of the macroscopic dielectric tensor
tensor_exc = mdf_file.get_tensor("exc")
tensor_exc.symmetrize(mdf_file.structure)
tensor_exc.plot(title="Si macroscopic dielectric tensor (Reduced coord)")

tensor_exc.plot(red_coords=False,
                title="Si macroscopic dielectric tensor (Cartesian coord)")

mdf_file.close()



"""
This module provides interfaces with the Materials Project REST API v2 to enable
the creation of data structures and pymatgen objects using Materials Project data.
"""
from __future__ import annotations

import sys
import pandas as pd

from collections import OrderedDict
from pprint import pprint
from monty.functools import lazy_property
from monty.string import marquee

from pymatgen.ext.matproj import MPRester, MPRestError
from abipy.tools.printing import print_dataframe
from abipy.core.mixins import NotebookWriter



MP_KEYS_FOR_DATAFRAME = (
    "pretty_formula", "e_above_hull", "energy_per_atom",
    "formation_energy_per_atom", "nsites", "volume",
    "spacegroup.symbol", "spacegroup.number",
    "band_gap", "total_magnetization", "material_id" # "unit_cell_formula", "icsd_id", "icsd_ids", "cif", , "tags", "elasticity")
)


def get_mprester():
    """
    Args:
    """
    rester = MPRester()
    #print(f"{type(rester)=}")
    return rester


#class MyMPRester(MPRester):
#    """
#    Subclass Materials project Rester.
#    See :cite:`Jain2013,Ong2015`.
#
#    .. rubric:: Inheritance Diagram
#    .. inheritance-diagram:: MyMPRester
#    """
#    Error = MPRestError
#
#    def get_phasediagram_results(self, elements) -> PhaseDiagramResults:
#        """
#        Contact the materials project database, fetch entries and build :class:``PhaseDiagramResults`` instance.
#
#        Args:
#            elements: List of chemical elements.
#        """
#        entries = self.get_entries_in_chemsys(elements, inc_structure="final")
#        return PhaseDiagramResults(entries)


#class PhaseDiagramResults:
#    """
#    Simplified interface to phase-diagram pymatgen API.
#
#    Inspired to:
#
#        https://anaconda.org/matsci/plotting-and-analyzing-a-phase-diagram-using-the-materials-api/notebook
#
#    See also: :cite:`Ong2008,Ong2010`
#    """
#    def __init__(self, entries):
#        self.entries = entries
#        # Convert pymatgen structure to Abipy.
#        from abipy.core.structure import Structure
#        for e in entries:
#            e.structure.__class__ = Structure
#
#        self.structures = [e.structure for e in entries]
#        self.mpids = [e.entry_id for e in entries]
#
#        # Create phase diagram.
#        from pymatgen.analysis.phase_diagram import PhaseDiagram
#        self.phasediagram = PhaseDiagram(self.entries)
#
#    def plot(self, show_unstable=True, show=True):
#        """
#        Plot phase diagram.
#
#        Args:
#            show_unstable (float): Whether unstable phases will be plotted as
#                well as red crosses. If a number > 0 is entered, all phases with
#                ehull < show_unstable will be shown.
#            show: True to show plot.
#
#        Return: plotter object.
#        """
#        from pymatgen.analysis.phase_diagram import PDPlotter
#        plotter = PDPlotter(self.phasediagram, show_unstable=show_unstable)
#        if show:
#            plotter.show()
#        return plotter
#
#    @lazy_property
#    def dataframe(self) -> pd.DataFrame:
#        """Pandas dataframe with the most important results."""
#        rows = []
#        for e in self.entries:
#            d = e.structure.get_dict4pandas(with_spglib=True)
#            decomp, ehull = self.phasediagram.get_decomp_and_e_above_hull(e)
#
#            rows.append(OrderedDict([
#                ("Materials ID", e.entry_id),
#                ("spglib_symb", d["spglib_symb"]), ("spglib_num", d["spglib_num"]),
#                ("Composition", e.composition.reduced_formula),
#                ("Ehull", ehull), # ("Equilibrium_reaction_energy", pda.get_equilibrium_reaction_energy(e)),
#                ("Decomposition", " + ".join(["%.2f %s" % (v, k.composition.formula) for k, v in decomp.items()])),
#            ]))
#
#        import pandas as pd
#        return pd.DataFrame(rows, columns=list(rows[0].keys()) if rows else None)
#
#    def print_dataframes(self, with_spglib=False, file=sys.stdout, verbose=0) -> None:
#        """
#        Print pandas dataframe to file `file`.
#
#        Args:
#            with_spglib: True to compute spacegroup with spglib.
#            file: Output stream.
#            verbose: Verbosity level.
#        """
#        print_dataframe(self.dataframe, file=file)
#        if verbose:
#            from abipy.core.structure import dataframes_from_structures
#            dfs = dataframes_from_structures(self.structures, index=self.mpids, with_spglib=with_spglib)
#            print_dataframe(dfs.lattice, title="Lattice parameters:", file=file)
#            if verbose > 1:
#                print_dataframe(dfs.coords, title="Atomic positions (columns give the site index):", file=file)


class DatabaseStructures(NotebookWriter):
    """
    Store the results of a query to the MP database.
    This object is immutable, use add_entry to create a new instance.
    """

    def __init__(self, structures, ids, data=None):
        """
        Args:
            structures: List of structure objects
            ids: List of database ids.
            data: List of dictionaries with data associated to the structures (optional).
        """
        from abipy.core.structure import Structure
        self.structures = list(map(Structure.as_structure, structures))
        self.ids, self.data = ids, data
        assert len(self.structures) == len(ids)
        if data is not None:
            assert len(self.structures) == len(data)

    def __bool__(self):
        """bool(self)"""
        return bool(self.structures)

    __nonzero__ = __bool__  # py2

    def filter_by_spgnum(self, spgnum: int) -> MpStructures:
        """Filter structures by space group number. Return new MpStructures object."""
        inds = [i for i, s in enumerate(self.structures) if s.get_space_group_info()[1] == int(spgnum)]
        new_data = None if self.data is None else [self.data[i] for i in inds]
        return self.__class__([self.structures[i] for i in inds], [self.ids[i] for i in inds], data=new_data)

    def add_entry(self, structure, entry_id, data_dict=None):
        """
        Add new entry, return new object.

        Args:
           structure: New structure object.
           entry_id: ID associated to new structure.
           data_dict: Option dictionary with metadata.
        """
        if data_dict is None:
            new_data = None if self.data is None else self.data + [{}]
        else:
            assert self.data is not None
            new_data = self.data + [data_dict]

        return self.__class__(self.structures + [structure], self.ids + [entry_id], data=new_data)

    @property
    def lattice_dataframe(self) -> pd.DataFrame:
        """pandas DataFrame with lattice parameters."""
        return self.structure_dataframes.lattice

    @property
    def coords_dataframe(self) -> pd.DataFrame:
        """pandas DataFrame with atomic positions."""
        return self.structure_dataframes.coords

    @lazy_property
    def structure_dataframes(self):
        """Pandas dataframes constructed from self.structures."""
        from abipy.core.structure import dataframes_from_structures
        return dataframes_from_structures(self.structures, index=self.ids, with_spglib=True)

    def print_results(self, fmt="abivars", verbose=0, file=sys.stdout) -> None:
        """
        Print pandas dataframe, structures using format `fmt`, and data to file `file`.
        `fmt` is automaticall set to `cif` if structure is disordered.
        Set fmt to None or empty string to disable structure output.
        """
        print("\n# Found %s structures in %s database (use `verbose` to get further info)\n"
              % (len(self.structures), self.dbname), file=file)

        if self.dataframe is not None: print_dataframe(self.dataframe, file=file)
        if verbose and self.data is not None: pprint(self.data, stream=file)

        # Print structures
        print_structures = not (fmt is None or str(fmt) == "None")
        if print_structures:
            for i, structure in enumerate(self.structures):
                print(" ", file=file)
                print(marquee("%s input for %s" % (fmt, self.ids[i]), mark="#"), file=file)
                print("# " + structure.spget_summary(verbose=verbose).replace("\n", "\n# ") + "\n", file=file)
                if not structure.is_ordered:
                    print(structure.convert(fmt="cif"), file=file)
                else:
                    print(structure.convert(fmt=fmt), file=file)
                print(2 * "\n", file=file)

        if len(self.structures) > 10:
            # Print info again
            print("\n# Found %s structures in %s database (use `verbose` to get further info)\n"
                  % (len(self.structures), self.dbname), file=file)

    def yield_figs(self, **kwargs):  # pragma: no cover
        """NOP required by NotebookWriter protocol."""
        yield None

    def write_notebook(self, nbpath=None, title=None) -> str:
        """
        Write a jupyter notebook to nbpath. If nbpath is None, a temporay file in the current
        working directory is created. Return path to the notebook.
        """
        nbformat, nbv, nb = self.get_nbformat_nbv_nb(title=title)

        # Use pickle files for data persistence.
        tmpfile = self.pickle_dump()

        nb.cells.extend([
            #nbv.new_markdown_cell("# This is a markdown cell"),
            nbv.new_code_cell("dbs = abilab.restapi.DatabaseStructures.pickle_load('%s')" % tmpfile),
            nbv.new_code_cell("import qgrid"),
            nbv.new_code_cell("# dbs.print_results(fmt='cif', verbose=0)"),
            nbv.new_code_cell("# qgrid.show_grid(dbs.lattice_dataframe)"),
            nbv.new_code_cell("# qgrid.show_grid(dbs.coords_dataframe)"),
            nbv.new_code_cell("qgrid.show_grid(dbs.dataframe)"),
        ])

        return self._write_nb_nbpath(nb, nbpath)


class MpStructures(DatabaseStructures):
    """
    Store the results of a query to the Materials Project database.

    .. inheritance-diagram:: MpStructures
    """
    dbname = "Materials Project"

    @lazy_property
    def dataframe(self) -> pd.DataFrame:
        """
        Pandas dataframe constructed from self.data. None if data is not available.
        """
        if not self.data: return None
        import pandas as pd
        rows = []
        for d, structure in zip(self.data, self.structures):
            d = Dotdict(d)
            d = OrderedDict([(k, d.dotget(k, default=None)) for k in MP_KEYS_FOR_DATAFRAME])
            # Add lattice parameters.
            for k in ("a", "b", "c", "alpha", "beta", "gamma"):
                d[k] = getattr(structure.lattice, k)
            rows.append(d)

        return pd.DataFrame(rows, index=self.ids, columns=list(rows[0].keys()))

    def open_browser(self, browser=None, limit=10):
        """
        Args:
            browser: Open webpage in ``browser``. Use default if $BROWSER if None.
            limit: Max number of tabs opened in browser. None for no limit.
        """
        import webbrowser
        import cgi
        for i, mpid in enumerate(self.ids):
            if limit is not None and i >= limit:
                print("Found %d structures found. Won't open more than %d tabs" % (len(self.ids), limit))
                break
            # https://materialsproject.org/materials/mp-2172/
            url = "https://materialsproject.org/materials/%s/" % mpid
            webbrowser.get(browser).open_new_tab(cgi.escape(url))


class CodStructures(DatabaseStructures):
    """
    Store the results of a query to the COD_ database :cite:`Grazulis2011`.

    .. inheritance-diagram:: CodStructures
    """
    dbname = "COD"

    @lazy_property
    def dataframe(self) -> pd.DataFrame:
        """
        |pandas-Dataframe| constructed. Essentially geometrical info and space groups found by spglib_
        as COD API is rather limited.
        """
        df = self.lattice_dataframe.copy()
        # Add space group from COD
        df["cod_sg"] = [d.get("sg", "").replace(" ", "") for d in self.data]
        return df


class Dotdict(dict):

    def dotget(self, key, default=None):
        """
        d.dotget["foo.bar"] --> d["foo"]["bar"] if "foo.bar" not in self
        """
        # if key is in dict access as normal
        if key in self:
            return super().__getitem__(key)

        # Assume string
        i = -1
        try:
            i = key.find(".")
            if i == -1: return default
        except AttributeError:
            return default

        try:
            root, key = key[:i], key[i+1:]
            if key == ".": return None
            return Dotdict(**self[root])[key]
        except Exception:
            return None



#!/usr/bin/env python
r"""
e-Bands with frozen phonon
==========================

Electronic band structure of silicon in a distorted geometry (frozen phonon at q=0)
"""

import sys
import os
import numpy as np
import abipy.data as data
import abipy.abilab as abilab
import abipy.flowtk as flowtk


def make_scf_nscf_inputs(structure, paral_kgb=1):
    multi = abilab.MultiDataset(structure, pseudos=data.pseudos("14si.pspnc"), ndtset=2)

    # Global variables
    global_vars = dict(
        ecut=6,
        nband=8,
        timopt=-1,
        paral_kgb=0,
        #nstep=4, # This is not enough to converge. Used to test the automatic restart.
        nstep=10,
        iomode=3,
    )

    multi.set_vars(global_vars)

    # Dataset 1 (GS run)
    multi[0].set_kmesh(ngkpt=[8,8,8], shiftk=[0,0,0])
    multi[0].set_vars(tolvrs=1e-6)

    # Dataset 2 (NSCF run)
    kptbounds = [
        [0.5, 0.0, 0.0], # L point
        [0.0, 0.0, 0.0], # Gamma point
        [0.0, 0.5, 0.5], # X point
    ]

    multi[1].set_kpath(ndivsm=6, kptbounds=kptbounds)
    multi[1].set_vars(tolwfr=1e-12)

    # Generate two input files for the GS and the NSCF run
    scf_input, nscf_input = multi.split_datasets()

    return scf_input, nscf_input


def build_flow(options):
    # Working directory (default is the name of the script with '.py' removed and "run_" replaced by "flow_")
    if not options.workdir:
        options.workdir = os.path.basename(sys.argv[0]).replace(".py", "").replace("run_", "flow_")

    # build the structures
    base_structure = abilab.Structure.from_file(data.cif_file("si.cif"))
    modifier = abilab.StructureModifier(base_structure)

    etas = [-0.1, 0, +0.1]
    ph_displ = np.reshape(np.zeros(3*len(base_structure)), (-1,3))
    ph_displ[0,:] = [+1, 0, 0]
    ph_displ[1,:] = [-1, 0, 0]

    displaced_structures = modifier.displace(ph_displ, etas, frac_coords=False)

    flow = flowtk.Flow(options.workdir, manager=options.manager)

    for structure in displaced_structures:
        # Create the work for the band structure calculation.
        scf_input, nscf_input = make_scf_nscf_inputs(structure)

        work = flowtk.BandStructureWork(scf_input, nscf_input)
        flow.register_work(work)

    return flow


# This block generates the thumbnails in the AbiPy gallery.
# You can safely REMOVE this part if you are using this script for production runs.
if os.getenv("READTHEDOCS", False):
    __name__ = None
    import tempfile
    options = flowtk.build_flow_main_parser().parse_args(["-w", tempfile.mkdtemp()])
    build_flow(options).graphviz_imshow()


@flowtk.flow_main
def main(options):
    """
    This is our main function that will be invoked by the script.
    flow_main is a decorator implementing the command line interface.
    Command line args are stored in `options`.
    """
    return build_flow(options)


if __name__ == "__main__":
    sys.exit(main())


############################################################################
#
# Run the script with:
#
#     run_phfrozen_ebands -s
#
# then use:
#
#    abirun.py flow_phfrozen_ebands/ structures -v
#
# to analyze the input/output structures including the atomic positions:
#
# .. code-block:: bash
#
#       Lattice parameters:
#                 formula  natom  angle0  angle1  angle2      a      b      c  volume  \
#       w0_t0_in      Si2      2    60.0    60.0    60.0  3.867  3.867  3.867  40.888
#       w0_t0_out     Si2      2    60.0    60.0    60.0  3.867  3.867  3.867  40.888
#       w0_t1_in      Si2      2    60.0    60.0    60.0  3.867  3.867  3.867  40.888
#       w1_t0_in      Si2      2    60.0    60.0    60.0  3.867  3.867  3.867  40.888
#       w1_t0_out     Si2      2    60.0    60.0    60.0  3.867  3.867  3.867  40.888
#       w1_t1_in      Si2      2    60.0    60.0    60.0  3.867  3.867  3.867  40.888
#       w2_t0_in      Si2      2    60.0    60.0    60.0  3.867  3.867  3.867  40.888
#       w2_t0_out     Si2      2    60.0    60.0    60.0  3.867  3.867  3.867  40.888
#       w2_t1_in      Si2      2    60.0    60.0    60.0  3.867  3.867  3.867  40.888
#
#                 abispg_num  P [GPa]  Max|F| eV/ang task_class              status
#       w0_t0_in        None      NaN            NaN    ScfTask  Completed
#       w0_t0_out         12   -3.638      3.180e+00    ScfTask  Completed
#       w0_t1_in        None      NaN            NaN   NscfTask  Completed
#       w1_t0_in        None      NaN            NaN    ScfTask  Completed
#       w1_t0_out        227   -5.212      7.430e-27    ScfTask  Completed
#       w1_t1_in        None      NaN            NaN   NscfTask  Completed
#       w2_t0_in        None      NaN            NaN    ScfTask  Completed
#       w2_t0_out         12   -4.192      2.095e+00    ScfTask  Completed
#       w2_t1_in        None      NaN            NaN   NscfTask  Completed
#
#       Atomic positions (columns give the site index):
#                                                    0  \
#       w0_t0_in   (Si, +0.970139 +0.000000 +0.014930)
#       w0_t0_out  (Si, +0.970139 +0.000000 +0.014930)
#       w0_t1_in   (Si, +0.970139 +0.000000 +0.014930)
#       w1_t0_in   (Si, +0.000000 +0.000000 +0.000000)
#       w1_t0_out  (Si, +0.000000 +0.000000 +0.000000)
#       w1_t1_in   (Si, +0.000000 +0.000000 +0.000000)
#       w2_t0_in   (Si, +0.029861 +0.000000 +0.985070)
#       w2_t0_out  (Si, +0.029861 +0.000000 +0.985070)
#       w2_t1_in   (Si, +0.029861 +0.000000 +0.985070)
#
#                                                    1              status
#       w0_t0_in   (Si, +0.279861 +0.250000 +0.235070)  Completed
#       w0_t0_out  (Si, +0.279861 +0.250000 +0.235070)  Completed
#       w0_t1_in   (Si, +0.279861 +0.250000 +0.235070)  Completed
#       w1_t0_in   (Si, +0.250000 +0.250000 +0.250000)  Completed
#       w1_t0_out  (Si, +0.250000 +0.250000 +0.250000)  Completed
#       w1_t1_in   (Si, +0.250000 +0.250000 +0.250000)  Completed
#       w2_t0_in   (Si, +0.220139 +0.250000 +0.264930)  Completed
#       w2_t0_out  (Si, +0.220139 +0.250000 +0.264930)  Completed
#       w2_t1_in   (Si, +0.220139 +0.250000 +0.264930)  Completed
#
#  Finally, we can plot the electronic bands with the command:
#
#    abirun.py flow_phfrozen_ebands ebands -p -t NscfTask
#
#  to select only the band structures produced by the NscfTask.
#
# .. image:: https://github.com/abinit/abipy_assets/blob/master/run_phfrozen_ebands.png?raw=true
#    :alt: Band structures of Si computed for different displacement amplitudes.
#



"""Configuration file for pytest."""
from __future__ import annotations

import os
import pytest
#import yaml
import ruamel.yaml as yaml
from abipy.tools.iotools import yaml_safe_load_path, yaml_dump
import copy
import abipy.flowtk as flowtk

from monty.collections import AttrDict
from monty.string import marquee

# Are we running on travis?
on_travis = os.environ.get("TRAVIS", "False") == "true"

# Create the list of Manager configurations used for the integration tests.
# Note that the items in _manager_confs must be hashable hence we cannot use dictionaries directly
# To bypass this problem we operate on dictionaries to generate the different configuration
# and then we convert the dictionary to string with yaml.dump. This string will be passed
# to Manager.from_string in fwp. base_conf looks like:

USER_CONFIG_DIR = os.path.join(os.path.expanduser("~"), ".abinit", "abipy")
#USER_CONFIG_DIR = os.path.dirname(__file__)

# Read the base configuration from file
#with open(os.path.join(USER_CONFIG_DIR, "manager.yml")) as fh:
#    base_conf = yaml.safe_load(fh)

base_conf = yaml_safe_load_path(os.path.join(USER_CONFIG_DIR, "manager.yml"))

# Build list of configurations.
_manager_confs = []

for autoparal in [1]: #, 1]:
    newd = copy.deepcopy(base_conf)
    if "policy" not in newd: newd["policy"] = {}
    newd["policy"]["autoparal"] = autoparal
    _manager_confs.append(newd)

_manager_confs = [yaml_dump(d) for d in _manager_confs]
#_manager_confs = [base_conf]


@pytest.fixture(params=_manager_confs)
def fwp(tmpdir, request):
    """
    Parameters used to initialize Flows.

    This fixture allows us to change the |TaskManager| so that we can easily test different configurations.
    """
    # Temporary working directory
    fwp.workdir = str(tmpdir)

    # Create the TaskManager.
    fwp.manager = flowtk.TaskManager.from_string(request.param)
    fwp.scheduler = flowtk.PyFlowScheduler.from_file(os.path.join(USER_CONFIG_DIR, "scheduler.yml"))
    fwp.on_travis = on_travis
    fwp.abinit_build = flowtk.AbinitBuild()

    return fwp


# Use tuples instead of dict because pytest require objects to be hashable.
if on_travis:
    _tvars_list = [
        (("paral_kgb", 0),),
        #(("paral_kgb", 1),),
    ]

else:
    _tvars_list = [
        #(("paral_kgb", 0),),
        (("paral_kgb", 1),),
    ]


@pytest.fixture(params=_tvars_list)
def tvars(request):
    """
    Abinit variables passed to the test functions.

    This fixture allows us change the variables in the input files
    so that we can easily test different scenarios e.g. runs with or without paral_kgb == 1
    """
    return AttrDict({k: v for k, v in request.param})


def pytest_addoption(parser):
    """Add extra command line options."""
    parser.addoption('--loglevel', default="ERROR", type=str,
                     help="Set the loglevel. Possible values: CRITICAL, ERROR (default), WARNING, INFO, DEBUG")
    #parser.addoption('--manager', default=None, help="TaskManager file (defaults to the manager.yml found in cwd"


def pytest_report_header(config):
    """Write the initial header."""
    lines = []
    app = lines.append
    app("\n" + marquee("Begin integration tests for AbiPy + abinit", mark="="))
    app("\tAssuming the environment is properly configured:")
    app("\tIn particular, the abinit executable must be in $PATH.")
    app("\tChange manager.yml according to your platform.")
    app("\tNumber of TaskManager configurations used: %d" % len(_manager_confs))

    if config.option.verbose > 0:
        for i, s in enumerate(_manager_confs):
            app(80 * "=")
            app("TaskManager #%d" % i)
            app(s)
            app(80 * "=")
    app("")

    # Print info on Abinit build
    abinit_build = flowtk.AbinitBuild()
    print()
    print(abinit_build)
    print()
    if not config.option.verbose:
        print("Use --verbose for additional info")
    else:
        print(abinit_build.info)

    # Initialize logging
    # loglevel is bound to the string value obtained from the command line argument.
    # Convert to upper case to allow the user to specify --loglevel=DEBUG or --loglevel=debug
    import logging
    numeric_level = getattr(logging, config.option.loglevel.upper(), None)
    if not isinstance(numeric_level, int):
        raise ValueError('Invalid log level: %s' % config.option.loglevel)
    logging.basicConfig(level=numeric_level)

    return lines



# coding: utf-8
"""This module contains the class describing a planewave wavefunction."""
from __future__ import annotations
#import copy
import numpy as np

from monty.termcolor import cprint
from abipy.core import Mesh3D
from abipy.core.structure import Structure
from abipy.core.kpoints import Kpoint
from abipy.iotools import Visualizer
from abipy.iotools.xsf import xsf_write_structure, xsf_write_data
from abipy.tools.plotting import add_fig_kwargs, get_ax_fig_plt, get_axarray_fig_plt
from abipy.tools.typing import Figure


__all__ = [
    "PWWaveFunction",
]


def latex_label_ispinor(ispinor: int, nspinor: int) -> str:
    if nspinor == 1:
        return ""
    elif nspinor == 2:
        return {k: v.replace("myuparrow", "uparrow") for k, v in
                    {0: r"$\sigma=\myuparrow$", 1: r"$\sigma=\downarrow$"}.items()}[ispinor]
    else:
        raise ValueError("Wrong value for nspinor: %s" % nspinor)


class WaveFunction:
    """
    Abstract class defining base and abstract methods for wavefunction objects.
    """
    def __eq__(self, other):
        if other is None: return False
        if self.gsphere != other.gsphere: return False
        return np.allclose(self.ug, other.ug)

    def __ne__(self, other):
        return not (self == other)

    def __iter__(self):
        """Yields G, ug[0:nspinor, G]"""
        return zip(self.gvecs, self.ug.T)

    def __getitem__(self, slice):
        return self.gvecs[slice], self.ug[:, slice]

    def __repr__(self):
        return str(self)

    def __str__(self):
        return self.to_string()

    @property
    def shape(self) -> tuple[int, int]:
        """Shape of ug i.e. (nspinor, npw)"""
        return self.nspinor, self.npw

    @property
    def gsphere(self):
        """:class:`GSphere` object"""
        return self._gsphere

    @property
    def kpoint(self):
        """|Kpoint| object"""
        return self.gsphere.kpoint

    @property
    def gvecs(self):
        """G-vectors in reduced coordinates."""
        return self.gsphere.gvecs

    @property
    def npw(self) -> int:
        """Number of G-vectors."""
        return len(self.gsphere)

    @property
    def ecut(self) -> float:
        """Cutoff energy in Hartree."""
        return self.gsphere.ecut

    @property
    def isnc(self) -> bool:
        """True if norm-conserving wavefunction."""
        return isinstance(self, PWWaveFunction)

    @property
    def ispaw(self) -> bool:
        """True if PAW wavefunction."""
        return isinstance(self, PAW_WaveFunction)

    @property
    def ug(self):
        """Periodic part of the wavefunctions in G-space."""
        return self._ug

    def set_ug(self, ug):
        """Set the value of the u(nspinor, G) array."""
        if ug.shape != self.shape:
            raise ValueError("Input ug shape %s differs from the one stored in self %s" % (ug.shape, self.shape))
        self._ug = ug
        self.delete_ur()

    @property
    def ur(self):
        """Periodic part of the wavefunction in real space."""
        try:
            return self._ur
        except AttributeError:
            self._ur = self.fft_ug()
            return self._ur

    def delete_ur(self) -> None:
        """Delete _u(r) (if it has been computed)."""
        try:
            del self._ur
        except AttributeError:
            pass

    #@property
    #def ur_xyz(self):
    #    """
    #    Returns a copy with ur[nspinor, nx, ny, nz]. Mainly used for post-processing.
    #    """
    #    return self.mesh.reshape(self.ur).copy()

    #@property
    #def ur2_xyz(self):
    #    """
    #    Returns ur2[nx, ny, nz]. Mainly used for post-processing.
    #    """
    #    return self.mesh.reshape(self.ur2)

    @property
    def mesh(self):
        """The mesh used for the FFT."""
        return self._mesh

    def set_mesh(self, mesh) -> None:
        """Change the FFT mesh. `u(r)` will be computed on this box."""
        assert isinstance(mesh, Mesh3D)
        self._mesh = mesh
        self.delete_ur()

    #def deepcopy(self):
    #    """Deep copy of self."""
    #    return copy.deepcopy(self)

    def get_ug_mesh(self, mesh=None):
        """
        Returns u(G) on the FFT mesh

        Args:
            mesh: |Mesh3d| object. If mesh is None, the internal mesh is used.
        """
        mesh = self.mesh if mesh is None else mesh
        return self.gsphere.tofftmesh(mesh, self.ug)

    def get_ur_mesh(self, mesh, copy=True):
        """
        Returns u(r) on the FFT mesh. This routine is mainly used if we need u(r) on a
        differen mesh. Data on the initial mesh is already available in `self.ur`

        Args:
            mesh: |Mesh3d| object.
            copy: By default, we return a copy of ur if mesh == self.mesh.
        """
        if mesh == self.mesh:
            return self.ur.copy() if copy else self.ur
        else:
            return self.fft_ug(mesh=mesh)

    def fft_ug(self, mesh=None):
        """
        Performs the FFT transform of :math:`u(g)` on mesh.

        Args:
            mesh: |Mesh3d| object. If mesh is None, self.mesh is used.

        Returns:
            :math:`u(r)` on the real space FFT box.
        """
        mesh = self.mesh if mesh is None else mesh
        ug_mesh = self.get_ug_mesh(mesh=mesh)
        return mesh.fft_g2r(ug_mesh, fg_ishifted=False)

    def to_string(self, verbose=0) -> str:
        """String representation."""
        lines = []; app = lines.append
        app("%s: nspinor: %d, spin: %d, band: %d " % (
            self.__class__.__name__, self.nspinor, self.spin, self.band))

        if hasattr(self, "gsphere"):
            app(self.gsphere.to_string(verbose=verbose))
        if hasattr(self, "mesh"):
            app(self.mesh.to_string(verbose=verbose))

        return "\n".join(lines)

    # TODO: get_ur2?
    @property
    def ur2(self):
        """
        [nx, ny, nz] array with :math:`||u(r)||^2` in real space.
        """
        ur2 = (self.ur.conj() * self.ur).real.copy()
        #if self.nspinor == 2: ur2 = ur2.sum(axis=3)
        return ur2


class PWWaveFunction(WaveFunction):
    """
    This object describes a wavefunction expressed in a plane-wave basis set.

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: PWWaveFunction
    """
    def __init__(self, structure: Structure, nspinor, spin, band, gsphere, ug):
        """
        Creation method.

        Args:
            structure: |Structure| object.
            nspinor: number of spinorial components.
            spin: spin index (only used if collinear-magnetism).
            band: band index (>=0)
            gsphere |GSphere| instance.
            ug: 2D array containing u[nspinor,G] for G in gsphere.
        """
        self.structure = structure
        self.nspinor, self.spin, self.band = nspinor, spin, band
        # Sanity check.
        assert ug.ndim == 2
        assert ug.shape[0] == nspinor
        assert ug.shape[1] == gsphere.npw

        self._gsphere = gsphere
        self._ug = np.array(ug)

    #def kinetic_energy(self):
    #    """Computes the matrix element of the kinetic operator in reciprocal space."""
    #    tug = -0.5 * self.gsphere.kpg2 * self.ug
    #    return np.vdot(self.ug, tug).sum()

    def norm2(self, space="g") -> float:
        r"""
        Return :math:`||\psi||^2` computed in G- or r-space.

        space:  Integration space. Possible values ["g", "gsphere", "r"]
            if "g" or "r" the scalar product is computed in G- or R-space on the FFT box.
            if "gsphere" the integration is done on the G-sphere.
        """
        space = space.lower()

        if space == "g":
            return np.real(np.vdot(self.ug, self.ug))
        elif space == "gsphere":
            return np.real(np.vdot(self.ug, self.ug))
        elif space == "r":
            return np.vdot(self.ur, self.ur) / self.mesh.size
        else:
            raise ValueError("Wrong space: %s" % str(space))

    def braket(self, other, space="g") -> complex:
        """
        Returns the scalar product <u1|u2> of the periodic part of two wavefunctions
        computed in G-space or r-space, depending on the value of space.
        Note that selection rules introduced by k-points is not taken into accout.

        Args:
            other: Other wave (right-hand side)
            space:  Integration space. Possible values ["g", "gsphere", "r"]
                if "g" or "r" the scalar product is computed in G- or R-space on the FFT box.
                if "gsphere" the integration is done on the G-sphere. Note that
                this option assumes that self and other have the same list of G-vectors.
        """
        space = space.lower()

        if space == "g":
            ug1_mesh = self.gsphere.tofftmesh(self.mesh, self.ug)
            ug2_mesh = other.gsphere.tofftmesh(self.mesh, other.ug) if other is not self else ug1_mesh
            return np.vdot(ug1_mesh, ug2_mesh)
        elif space == "gsphere":
            return np.vdot(self.ug, other.ug)
        elif space == "r":
            return np.vdot(self.ur, other.ur) / self.mesh.size
        else:
            raise ValueError("Wrong space: %s" % str(space))

    def get_interpolator(self):
        """
        Return an interpolator object that interpolates periodic functions in real space.
        """
        from abipy.tools.numtools import BlochRegularGridInterpolator
        return BlochRegularGridInterpolator(self.structure, self.ur)

    #def pww_translation(self, gvector, rprimd):
    #    """Returns the pwwave of the kpoint translated by one gvector."""
    #    gsph = self.gsphere.copy()
    #    wpww = PWWaveFunction(self.structure, self.nspinor, self.spin, self.band, gsph, self.ug.copy())
    #    wpww.mesh = self.mesh
    #    wpww.pww_translation_inplace(gvector, rprimd)
    #    return wpww

    #def pww_translation_inplace(self, gvector, rprimd):
    #    """Translates the pwwave from 1 kpoint by one gvector."""
    #    self.gsphere.kpoint = self.gsphere.kpoint + gvector
    #    self.gsphere.gvecs = self.gsphere.gvecs + gvector
    #    fft_ndivs = (self.mesh.shape[0] + 2, self.mesh.shape[1] + 2, self.mesh.shape[2] + 2)
    #    newmesh = Mesh3D(fft_ndivs, rprimd, pbc=True)
    #    self.mesh = newmesh

    #def pwwtows_inplace(self):
    #    """Wrap the kpoint to the interval ]-1/2,1/2] and update pwwave accordingly."""
    #    kpoint = Kpoint(self.gsphere.kpoint, self.gsphere.gprimd)
    #    wkpt = kpoint.wrap_to_ws()

    #    if np.allclose(wkpt.rcoord, kpoint.rcoord):
    #        return

    #    #@David FIXME this is wrong
    #    gvector = np.array(kpoint.rcoord - wkpt.rcoord, np.int)
    #    self.gsphere.gvecs = self.gsphere.gvecs + gvector
    #    self.gsphere.kpoint = wkpt.rcoord

    #def pwwtows(self):
    #    """Return the pwwave of the kpoint wrapped to the interval ]-1/2,1/2]."""
    #    gsph = self.gsphere.copy()
    #    wpww = PWWaveFunction(self.structure, self.nspinor, self.spin, self.band, gsph, self.ug.copy())
    #    wpww.pwwtows_inplace()
    #    wpww.mesh = self.mesh
    #    return wpww

    #def rotate(self, symmop, mesh=None):
    #    """
    #    Apply the symmetry operation `symmop` to the periodic part.

    #    Args:
    #        symmop: :class:`Symmetry` operation
    #        mesh: mesh for the FFT, if None the mesh of self is used.

    #    Returns:
    #        New wavefunction object.
    #    """
    #    if self.nspinor != 1:
    #        raise ValueError("Spinor rotation not available yet.")

    #    rot_gsphere = self.gsphere.rotate(symmop)
    #    #rot_istwfk = istwfk(rot_kpt)

    #    if not np.allclose(symmop.tau, np.zeros(3)):
    #        rot_ug = np.empty_like(self.ug)
    #        rot_gvecs = rot_gsphere.gvecs
    #        rot_kpt = rot_gsphere.kpoint.frac_coords

    #        ug = self._ug
    #        #phase = np.exp(-2j * np.pi * (np.dot(rot_gvecs + rot_kpt, symmop.tau)))
    #        for ig in range(self.npw):
    #            rot_ug[:, ig] = ug[:, ig] * np.exp(-2j * np.pi * (np.dot(rot_gvecs[ig] + rot_kpt, symmop.tau)))
    #    else:
    #        rot_ug = self.ug.copy()

    #    # Invert the collinear spin if we have an AFM operation.
    #    rot_spin = self.spin
    #    if self.nspinor == 1:
    #        rot_spin = self.spin if symmop.is_fm else (self.spin + 1) % 2

    #    # Build new wave and set the mesh.
    #    new = self.__class__(self.nspinor, rot_spin, self.band, rot_gsphere, rot_ug)
    #    new.set_mesh(mesh if mesh is not None else self.mesh)
    #    return new

    @add_fig_kwargs
    def plot_line(self, point1, point2, num=200, with_krphase=False, cartesian=False,
                  ax=None, fontsize=12, **kwargs) -> Figure:
        """
        Plot (interpolated) wavefunction in real space along a line defined by ``point1`` and ``point2``.

        Args:
            point1: First point of the line. Accepts 3d vector or integer.
                The vector is in reduced coordinates unless ``cartesian`` is True.
                If integer, the first point of the line is given by the i-th site of the structure
                e.g. ``point1=0, point2=1`` gives the line passing through the first two atoms.
            point2: Second point of the line. Same API as ``point1``.
            num: Number of points sampled along the line.
            with_krphase: True to include the :math:`e^{ikr}` phase-factor.
            cartesian: By default, ``point1`` and ``point1`` are interpreted as points in fractional
                coordinates (if not integers). Use True to pass points in cartesian coordinates.
            ax: |matplotlib-Axes| or None if a new figure should be created.
            fontsize: legend and title fontsize.

        Return: |matplotlib-Figure|
        """
        # Interpolate along line.
        interpolator = self.get_interpolator()
        r = interpolator.eval_line(point1, point2, num=num, cartesian=cartesian,
                                   kpoint=None if not with_krphase else self.kpoint)
        # Plot data.
        ax, fig, plt = get_ax_fig_plt(ax=ax)
        which = r"\psi(r)" if with_krphase else "u(r)"
        for ispinor in range(self.nspinor):
            spinor_label = latex_label_ispinor(ispinor, self.nspinor)
            ur = r.values[ispinor]
            ax.plot(r.dist, ur.real, label=r"$\Re %s$ %s" % (which, spinor_label))
            ax.plot(r.dist, ur.imag, label=r"$\Im %s$ %s" % (which, spinor_label))
            ax.plot(r.dist, ur.real**2 + ur.imag**2, label=r"$|\psi(r)|^2$ %s" % spinor_label)

        ax.grid(True)
        ax.set_xlabel("Distance from site1 [Angstrom]")
        ax.legend(loc="best", fontsize=fontsize, shadow=True)

        return fig

    @add_fig_kwargs
    def plot_line_neighbors(self, site_index, radius, num=200, with_krphase=False, 
                            max_nn=10, fontsize=12, **kwargs) -> Figure:
        """
        Plot (interpolated) density/potential in real space along the lines connecting
        an atom specified by ``site_index`` and all neighbors within a sphere of given ``radius``.

        .. warning:

            This routine can produce lots of plots! Be careful with the value of ``radius``.
            See also ``max_nn``.

        Args:
            site_index: Index of the atom in the structure.
            radius: Radius of the sphere in Angstrom.
            num: Number of points sampled along the line.
            with_krphase: True to include the :math:`e^{ikr}` phase-factor.
            max_nn: By default, only the first ``max_nn`` neighbors are showed.
            fontsize: legend and label fontsize.

        Return: |matplotlib-Figure|
        """
        site = self.structure[site_index]
        nn_list = self.structure.get_neighbors_old(site, radius, include_index=True)
        if not nn_list:
            cprint("Zero neighbors found for radius %s Ang. Returning None." % radius, "yellow")
            return None

        # Sort sites by distance.
        nn_list = list(sorted(nn_list, key=lambda t: t[1]))
        if max_nn is not None and len(nn_list) > max_nn:
            cprint("For radius %s, found %s neighbors but only max_nn %s sites are show." %
                   (radius, len(nn_list), max_nn), "yellow")
            nn_list = nn_list[:max_nn]

        # Get grid of axes (one row for neighbor)
        nrows, ncols = len(nn_list), 1
        ax_list, fig, plt = get_axarray_fig_plt(None, nrows=nrows, ncols=ncols,
                                                sharex=True, sharey=True, squeeze=True)
        ax_list = ax_list.ravel()

        interpolator = self.get_interpolator()
        kpoint = None if not with_krphase else self.kpoint
        which = r"\psi(r)" if with_krphase else "u(r)"

        # For each neighbor, plot psi along the line connecting site to nn.
        for i, (nn, ax) in enumerate(zip(nn_list, ax_list)):
            #nn_site, nn_dist, nn_sc_index = nn
            nn_site = nn
            nn_dist = nn.nn_distance
            nn_sc_index = nn.index
            title = "%s, %s, dist=%.3f A" % (nn_site.species_string, str(nn_site.frac_coords), nn_dist)

            r = interpolator.eval_line(site.frac_coords, nn_site.frac_coords, num=num, kpoint=kpoint)

            for ispinor in range(self.nspinor):
                spinor_label = latex_label_ispinor(ispinor, self.nspinor)
                ur = r.values[ispinor]
                ax.plot(r.dist, ur.real, label=r"$\Re %s$ %s" % (which, spinor_label))
                ax.plot(r.dist, ur.imag, label=r"$\Im %s$ %s" % (which, spinor_label))
                ax.plot(r.dist, ur.real**2 + ur.imag**2, label=r"$|\psi(r)|^2$ %s" % spinor_label)

            ax.set_title(title, fontsize=fontsize)
            ax.grid(True)

            if i == nrows - 1:
                ax.set_xlabel("Distance from site_index %s [Angstrom]" % site_index)
                ax.legend(loc="best", fontsize=fontsize, shadow=True)

        return fig

    def export_ur2(self, filename, visu=None):
        """
        Export :math:`|u(r)|^2` to file ``filename``.

        Args:
            filename: String specifying the file path and the file format.
                The format is defined by the file extension. filename="prefix.xsf", for example,
                will produce a file in XSF format. An *empty* prefix, e.g. ".xsf" makes the code use a temporary file.
            visu: :class:`Visualizer` subclass. By default, this method returns the first available
                visualizer that supports the given file format. If visu is not None, an
                instance of visu is returned. See :class:`Visualizer` for the list of
                applications and formats supported.

        Returns:
            Instance of :class:`Visualizer`
        """
        if "." not in filename:
            raise ValueError("Cannot detect file extension in: %s" % filename)

        tokens = filename.strip().split(".")
        ext = tokens[-1]

        if not tokens[0]:
            # fname == ".ext" ==> Create temporary file.
            # dir = os.getcwd() is needed when we invoke the method from a notebook.
            from abipy.core.globals import abinb_mkstemp
            _, filename = abinb_mkstemp(suffix="." + ext, text=True)
            print("Creating temporary file: %s" % filename)

        # Compute |u(r)|2 and write data according to ext.
        ur2 = np.reshape(self.ur2, (1,) + self.ur2.shape)

        with open(filename, mode="wt") as fh:
            if ext == "xsf":
                # xcrysden
                xsf_write_structure(fh, structures=self.structure)
                xsf_write_data(fh, self.structure, ur2, add_replicas=True)
            else:
                raise NotImplementedError("extension %s is not supported." % ext)

        if visu is None:
            return Visualizer.from_file(filename)
        else:
            return visu(filename)

    def visualize_ur2(self, appname="vesta"):
        """
        Visualize :math:`|u(r)|^2|`.

        See :class:`Visualizer` for the list of applications and formats supported.
        """
        # Get the Visualizer subclass from the string.
        visu = Visualizer.from_name(appname)

        # Try to export data to one of the formats supported by the visualizer
        # Use a temporary file (note "." + ext)
        for ext in visu.supported_extensions():
            ext = "." + ext
            try:
                return self.export_ur2(ext, visu=visu)
            except visu.Error:
                pass
        else:
            raise visu.Error("Don't know how to export data for %s" % str(appname))

    #def mvplot_cutplanes(self, show=True):
    #    data = self.ur2
    #    from abipy.display import mvtk
    #    figure, mlab = mvtk.get_fig_mlab(figure=None)
    #    source = mlab.pipeline.scalar_field(data)
    #    mlab.pipeline.image_plane_widget(source, plane_orientation='x_axes', slice_index=data.shape[0]//2)
    #    mlab.pipeline.image_plane_widget(source, plane_orientation='y_axes', slice_index=data.shape[1]//2)
    #    mlab.pipeline.image_plane_widget(source, plane_orientation='z_axes', slice_index=data.shape[2]//2)
    #    #mlab.pipeline.iso_surface(source, contours=contours) #, opacity=0.1)
    #    #mlab.pipeline.iso_surface(source, contours=[data.min()+ 0.1 * data.ptp()], opacity=0.1)
    #    mlab.outline()
    #    if show: mlab.show()
    #    return figure

    #def mvplot_volume(self, figure=None, vmin=0.65, vmax=0.9, show=True):
    #    from abipy.display import mvtk
    #    figure, mlab = mvtk.get_fig_mlab(figure=figure)
    #    data = self.ur2
    #    source = mlab.pipeline.scalar_field(data)
    #    data_min, data_max = data.min(), data.max()
    #    mlab.pipeline.volume(source,
    #                         vmin=data_min + vmin * (data_max - data_min),
    #                         vmax=data_min + vmax * (data_max - data_min))
    #    if show: mlab.show()
    #    return figure


class PAW_WaveFunction(WaveFunction):
    """
    All the methods that are related to the all-electron representation should start with ae.

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: PAW_WaveFunction
    """
